# FreeCoG Imaging Pipeline 
# Authors: Liberty Hamilton, Morgan Lee, David Chang, Zachary Greenberg
# Department of Neurological Surgery
# University of California, San Francisco
# Date Last Edited: March 9, 2017
#
# This file contains the Chang Lab imaging pipeline (freeCoG)
# as one importable python class for running a patients
# brain surface reconstruction and electrode localization/labeling

import os
import glob
import pickle

import nibabel as nib

import numpy as np
import scipy
import scipy.io
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot as plt
from matplotlib import cm
from matplotlib.backends.backend_pdf import PdfPages
from surface_warping_scripts.make_outer_surf import make_outer_surf # From ielu
from surface_warping_scripts.project_electrodes_anydirection import *

# For CT to MRI registration
from nipy.core.api import AffineTransform
import nipy.algorithms
import nipy.algorithms.resample
import nipy.algorithms.registration.histogram_registration

class freeCoG:
    ''' This defines the class freeCoG, which creates a patient object      
    for use in creating brain surface reconstructions, electrode placement,     
    and warping.        
    
    To initialize a patient, you must provide the subject ID, hemisphere,       
    freesurfer subjects directory, and (optionally) the freesurfer      
    executable directory. If these aren't specified they will default to the environment
    variables $SUBJECTS_DIR and $FREESURFER_HOME.
    
    For example:        
    
    >>> subj = 'EC1'     
    >>> subj_dir = '/usr/local/freesurfer/subjects'      
    >>> hem = 'rh'       
    >>> fs_dir = '/usr/local/freesurfer'     
    >>> patient = img_pipe.freeCoG(subj = subj, subj_dir = subj_dir, hem = hem, fs_dir = fs_dir)

    Attributes:         
        subj [str]: the subject ID      
        subj_dir [str]: the freesurfer subjects directory (e.g. /usr/local/freesurfer/subjects)     
        hem [str]: 'lh', 'rh', or 'stereo' -- the hemisphere of implantation        
        CT_dir [str]: the directory for CT imaging data (usually [subj_dir]/[subj]/CT)      
        elecs_dir [str]: the directory (usually [subj_dir]/[subj]/elecs)        
    '''

    def __init__(self, subj, hem, zero_indexed_electrodes=True, fs_dir=os.environ['FREESURFER_HOME'], subj_dir=os.environ['SUBJECTS_DIR']):
        '''
        subj: patient name (i.e. 'SUBJ_25')
        hem: patient hem of implantation ('lh' or 'rh')
        subj_dir: freesurfer subjects dir where you run img_pipe i.e. '/data_store2/imaging/subjects/'              
        fs_dir: the freesurfer executable directory (default: '/Applications/freesurfer')
        '''
        
        self.subj = subj
        self.subj_dir = subj_dir
        self.hem = hem
        self.img_pipe_dir = os.path.dirname(os.path.realpath(__file__))
        self.zero_indexed_electrodes = True

        #Freesurfer home directory
        self.fs_dir = fs_dir

        # CT_dir: dir for CT img data
        self.CT_dir = os.path.join(self.subj_dir, self.subj, 'CT')

        # elecs_dir: dir for elecs coordinates
        self.elecs_dir = os.path.join(self.subj_dir, self.subj, 'elecs')

        # Meshes directory for matlab/python meshes
        self.mesh_dir = os.path.join(self.subj_dir, self.subj, 'Meshes')
        surf_file = os.path.join(self.subj_dir, self.subj, 'Meshes', self.hem+'_pial_trivert.mat')
        if os.path.isfile(surf_file):
            self.pial_surf_file = dict()
            self.pial_surf_file['lh'] = os.path.join(self.subj_dir, self.subj, 'Meshes', 'lh_pial_trivert.mat')
            self.pial_surf_file['rh'] = os.path.join(self.subj_dir, self.subj, 'Meshes', 'rh_pial_trivert.mat')

        # surf directory
        self.surf_dir = os.path.join(self.subj_dir, self.subj, 'surf')

        # mri directory
        self.mri_dir = os.path.join(self.subj_dir, self.subj, 'mri')

        #if paths are not the default paths in the shell environment:
        os.environ['FREESURFER_HOME'] = fs_dir
        os.environ['SUBJECTS_DIR'] = subj_dir

    def prep_recon(self):
        '''Prepares file directory structure of subj_dir, copies acpc-aligned               
        T1.nii to the 'orig' directory and converts to mgz format.'''       

        # navigate to directory with subject freesurfer folders           
        # and make a new folder for the patient

        if not os.path.isdir(os.path.join(self.subj_dir, self.subj)):
            print("Making subject directory")
            os.mkdir(self.subj)

        # create elecs folders
        if not os.path.isdir(self.elecs_dir):
            print("Making electrodes directory")
            os.mkdir(self.elecs_dir)

        # create mri and orig folders
        if not os.path.isdir(self.mri_dir):
            os.mkdir(self.mri_dir)

        orig_dir = os.path.join(self.subj_dir, self.subj, 'mri', 'orig')
        if not os.path.isdir(orig_dir):
            os.mkdir(orig_dir)

        T1_file = os.path.join(self.subj_dir, self.subj, 'acpc', 'T1.nii')
        os.system('cp %s %s' %(T1_file, orig_dir))

        # convert T1 to freesurfer 001.mgz format
        T1_file2 = os.path.join(orig_dir, 'T1.nii')
        T1_mgz = os.path.join(orig_dir, '001.mgz')
        os.system('mri_convert %s %s'%(T1_file2, T1_mgz))

    def get_recon(self, flag_T3 = '-3T', openmp_flag='-openmp 12', gpu_flag=''):        
        '''Runs freesurfer recon-all for surface reconstruction.                
        Use flag_T3 = '-3T' if using scans from a 3T scanner, otherwise set flag_T3=''             
        openmp_flag = '-openmp 12' uses 12 cores and multithreading         
        to make recon-all faster. Otherwise use openmp_flag = ''        
        gpu_flag = '-use-gpu' if you want to run code on the GPU (some steps run faster)        
            otherwise use gpu_flag='' '''       
        os.system('recon-all -subjid %s -sd %s -all %s %s %s' % (self.subj, self.subj_dir, flag_T3, openmp_flag, gpu_flag))

        self.pial_surf_file = dict()
        self.pial_surf_file['lh'] = os.path.join(self.subj_dir, self.subj, 'Meshes', 'lh_pial_trivert.mat')
        self.pial_surf_file['rh'] = os.path.join(self.subj_dir, self.subj, 'Meshes', 'rh_pial_trivert.mat')

        #create gyri labels directory
        gyri_labels_dir = os.path.join(self.subj_dir, self.subj, 'label', 'gyri')
        if not os.path.isdir(gyri_labels_dir):
            os.mkdir(gyri_labels_dir)
            # This version of mri_annotation2label uses the coarse labels from the Desikan-Killiany Atlas
            os.system('mri_annotation2label --subject %s --hemi %s --surface pial --outdir %s'\
                %(self.subj, self.hem, gyri_labels_dir))

    def check_pial(self):
        '''Opens Freeview with the orig.mgz MRI loaded along with the pial surface. 
        User should scroll through to check that the pial surface corresponds correctly
        to the MRI.'''
        brain_mri = os.path.join(self.subj_dir, self.subj, 'mri', 'brain.mgz')
        lh_pial = os.path.join(self.subj_dir, self.subj, 'surf', 'lh.pial')
        rh_pial = os.path.join(self.subj_dir, self.subj, 'surf', 'rh.pial')
        os.system("freeview --volume %s --surface %s --surface %s --viewport 'coronal'" % (brain_mri, lh_pial, rh_pial))

    def make_dural_surf(self, radius=3, num_iter=30, dilate=0.0):
        '''
        Create smoothed dural surface for projecting electrodes to.
        '''
        # Create mask of pial surface
        hems = ['lh', 'rh']
        for hem in hems:
            print("Creating mask of %s pial surface"%(hem))
            pial_surf = os.path.join(self.subj_dir, self.subj, 'surf', hem+'.pial')
            pial_fill_image = os.path.join(self.subj_dir, self.subj, 'surf', hem+'.pial.filled.mgz')
            if not os.path.isfile(pial_fill_image):
                os.system('mris_fill -c -r 1 %s %s'%(pial_surf, pial_fill_image))
            outfile = os.path.join(self.subj_dir, self.subj, 'surf', hem+'.pial.outer')

            if os.path.isfile(pial_fill_image):
                # Create outer surface of this pial surface
                print("Creating outer surface for the filled pial surface %s"%(pial_fill_image))
                make_outer_surf(pial_surf, pial_fill_image, radius, outfile) # this is from ielu

                os.system('mris_extract_main_component %s %s-main'%(outfile, outfile))
                dura_surf = os.path.join(self.subj_dir, self.subj, 'surf', hem+'.dural')
                os.system('mris_smooth -nw -n %d %s-main %s'%(num_iter, outfile, dura_surf))

                if dilate != 0:
                    print("Dilating surface by %d mm"%(dilate))
                    if dilate > 0:
                        print("Multiplying dilate value by -1 to get outward dilation")
                        dilate = -1*dilate
                    os.system('mris_expand %s %d %s'%(dura_surf, dilate, dura_surf))
            else:
                print("Failed to create %s, check inputs."%(pial_fill_image))
        
        self.convert_fsmesh2mlab(mesh_name = 'dural')

    def mark_electrodes(self):
        ''' Launch the electrode picker for this subject. The electrode
        picker requires the Qt4Agg backend, so is launched via an external
        python script. 

        Inputs to the electrode_picker.py script include the subject directory and the hemisphere
        of implantation.
        '''
        individual_elecs_dir = os.path.join(self.subj_dir,self.subj,'elecs','individual_elecs')
        if not os.path.isdir(individual_elecs_dir):
            print("Creating directory %s"%(individual_elecs_dir))
            os.mkdir(individual_elecs_dir)
        print("Launching electrode picker")
        epicker = os.path.join(self.img_pipe_dir, 'SupplementalScripts', 'electrode_picker.py')
        os.system('python %s %s %s'%(epicker, os.path.join(self.subj_dir, self.subj), self.hem))

    def convert_fsmesh2mlab(self, mesh_name='pial'):
        '''Creates surface mesh triangle and vertex .mat files
        If no argument for mesh_name is given, lh.pial and rh.pial
        are converted into lh_pial_trivert.mat and rh_pial_trivert.mat
        in the Meshes directory (for use in python) and *_lh_pial.mat
        and *_rh_pial.mat for use in MATLAB.
        Other potential mesh_name values could be 'white' or 'inflated'.'''

        hems = ['lh', 'rh']

        if not os.path.isdir(self.mesh_dir):
            print('Making Meshes Directory')
            # Make the Meshes directory in subj_dir if it does not yet exist
            os.mkdir(self.mesh_dir)

        # Loop through hemispheres for this mesh, create one .mat file for each
        for h in hems:
            print("Making %s mesh"%(h))
            mesh_surf = os.path.join(self.surf_dir, h+'.'+mesh_name)
            vert, tri = nib.freesurfer.read_geometry(mesh_surf)
            out_file = os.path.join(self.mesh_dir, '%s_%s_trivert.mat'%(h, mesh_name))
            out_file_struct = os.path.join(self.mesh_dir, '%s_%s_%s.mat'%(self.subj, h, mesh_name))
            scipy.io.savemat(out_file, {'tri': tri, 'vert': vert})

            cortex = {'tri': tri+1, 'vert': vert}
            scipy.io.savemat(out_file_struct, {'cortex': cortex})

        setattr(self, mesh_name+'_surf_file', out_file)
  
    def reg_img(self, source='CT.nii', target='orig.mgz'):
        '''Runs nmi coregistration between two images.
        Usually run as patient.reg_img('CT.nii','orig.mgz').'''

        source_file = os.path.join(self.CT_dir, source)
        target_file = os.path.join(self.mri_dir, target)

        print("Computing registration from %s to %s"%(source_file, target_file))
        ctimg  = nipy.load_image(source_file)
        mriimg = nipy.load_image(target_file)

        ct_cmap = ctimg.coordmap  
        mri_cmap = mriimg.coordmap

        # Compute registration
        ct_to_mri_reg = nipy.algorithms.registration.histogram_registration.HistogramRegistration(ctimg, mriimg, similarity='nmi')
        aff = ct_to_mri_reg.optimize('affine').as_affine()   

        ct_to_mri = AffineTransform(ct_cmap.function_range, mri_cmap.function_range, aff)  
        reg_CT = nipy.algorithms.resample.resample(ctimg, mri_cmap, ct_to_mri.inverse(), mriimg.shape)    

        outfile = os.path.join(self.CT_dir, 'r'+source)
        print("Saving registered CT image as %s"%(outfile))
        nipy.save_image(reg_CT, outfile)

    def interp_grid(self, nchans = 256, grid_basename='hd_grid'):
        '''Interpolates corners for an electrode grid
        given the four corners (in order, 1, 16, 241, 256), or for
        32 channel grid, 1, 8, 25, 32.'''

        corner_file = os.path.join(self.elecs_dir, 'individual_elecs', grid_basename+'_corners.mat')
        corners = scipy.io.loadmat(corner_file)['elecmatrix']
        elecmatrix = np.zeros((nchans, 3))
        #you can add your own grid dimensions and corner indices here, if needed
        if nchans == 256:
            corner_nums = [0, 15, 240, 255]
            nrows = 16
            ncols = 16
        elif nchans == 20:
            corner_nums = [0,4,15,19]
            nrows = 5
            ncols = 4 
        elif nchans == 32:
            corner_nums = [0, 7, 24, 31]
            nrows = 8
            ncols = 4
        elif nchans == 64:
            corner_nums = [0, 7, 56, 63]
            nrows = 8
            ncols = 8
        elif nchans == 128:
            corner_nums = [0,15,112,127]
            nrows = 16 
            ncols= 8
        elif nchans == 56:
            corner_nums = [0, 6, 49, 55]
            nrows = 7
            ncols = 8

        # Add the electrode coordinates for the corners
        for i in np.arange(4):
            elecmatrix[corner_nums[i],:] = corners[i,:]

        # Interpolate over one dimension (vertical columns from corner 1 to 2 and corner 3 to 4)
        # loop over x, y, and z coordinates
        for i in np.arange(3):
            elecmatrix[corner_nums[0]:corner_nums[1]+1,i] = np.linspace(elecmatrix[corner_nums[0],i], elecmatrix[corner_nums[1], i], nrows)
            elecmatrix[corner_nums[2]:corner_nums[3]+1,i] = np.linspace(elecmatrix[corner_nums[2],i], elecmatrix[corner_nums[3], i], nrows)

        grid = np.arange(nchans).reshape((nrows, ncols), order='F')

        # Now fill in the rows using the new data from the leftmost and rightmost columns
        for row in np.arange(nrows):
            for i in np.arange(3):
                elecmatrix[grid[row,:],i] = np.linspace(elecmatrix[row,i], elecmatrix[row+(grid[0,-1]-grid[0,0]),i], ncols)

        orig_file = os.path.join(self.elecs_dir, 'individual_elecs', '%s_orig.mat'%(grid_basename))
        scipy.io.savemat(orig_file, {'elecmatrix': elecmatrix} )

    def project_electrodes(self, elecfile_prefix='hd_grid', use_mean_normal=True, \
                                 surf_type='dural', \
                                 num_iter=30, dilate=0.0, grid=True):
        '''elecfile_prefix: prefix of the .mat file with the electrode coordinates matrix 
        use_mean_normal: whether to use mean normal vector (mean of the 4 normal vectors from the grid's 
        corner electrodes) as the projection direction
        surf_type: 'dural' or 'pial'
        projection_method: 'convex_hull','none','alphavol'
        num_iter: how many smoothing iterations when creating the dural surface
        
        Projects the electrodes of a grid based on the mean normal vector of the four grid
        corner electrodes that were manually localized from the registered CT.'''

        print('Projection Params: \n\t Grid Name: %s.mat \n\t Use Mean Normal: %s \n\t \
               Surface Type: %s \n\t Number of Smoothing Iterations (if using dural): %d'\
                %(elecfile_prefix,use_mean_normal,surf_type,num_iter))
        if grid: 
            corners_file = os.path.join(self.elecs_dir, 'individual_elecs', elecfile_prefix+'_corners.mat')
            elec_corners = scipy.io.loadmat(corners_file)['elecmatrix']
            elecfile_name = elecfile_prefix +'_orig'
            if not os.path.isdir(os.path.join(self.elecs_dir, 'individual_elecs', 'preproc')):
                print('Making preproc directory')
                os.mkdir(os.path.join(self.elecs_dir, 'individual_elecs','preproc'))
        else:
            elecfile_name = elecfile_prefix

        dural_mesh = os.path.join(self.subj_dir, self.subj, 'Meshes', self.subj + '_' + self.hem + '_dural.mat')
        if surf_type=='dural' and not os.path.isfile(dural_mesh):
            print('Creating dural surface mesh, using %d smoothing iterations'%(num_iter))
            self.make_dural_surf(num_iter=num_iter, dilate=dilate)

        if use_mean_normal and grid:
            #the corners draw out a 'rectangle', each side is a grid vector
            grid_vectors = elec_corners[1] - elec_corners[0], \
                           elec_corners[3] - elec_corners[2], \
                           elec_corners[3] - elec_corners[1], \
                           elec_corners[2] - elec_corners[0]
            #we can get 4 normal vectors from the 4 grid vectors 
            normal_vectors = [np.cross(grid_vectors[0],grid_vectors[2]), \
                              np.cross(grid_vectors[0],grid_vectors[3]), \
                              np.cross(grid_vectors[1],grid_vectors[2]), \
                              np.cross(grid_vectors[1],grid_vectors[3])]
            
            #normalize the normal vectors
            for vec in range(4):
                normal_vectors[vec] = normal_vectors[vec]/np.linalg.norm(normal_vectors[vec])

            #get all the normal vectors facing the same direction
            for vec in range(1,4):
                angle = np.arccos(np.dot(normal_vectors[vec],normal_vectors[vec-1])/(np.linalg.norm(normal_vectors[vec])*np.linalg.norm(normal_vectors[vec-1])))
                #flip the normal vector and see if angle between it and the previous normal vector is smaller
                flipped_angle = np.arccos(np.dot((-1.0*normal_vectors[vec]),normal_vectors[vec-1])/(np.linalg.norm(normal_vectors[vec])*np.linalg.norm(normal_vectors[vec-1])))
                if angle > flipped_angle:
                    normal_vectors[vec] = normal_vectors[vec]*-1.0
            print('Normal vectors: ' + str(normal_vectors))

            #take the mean of the 4 normal vectors
            mean_normal = np.array(normal_vectors).mean(axis=0)

            #mean normal vector should point outwards (laterally, not towards the center of the brain)
            #if it doesn't, flip the mean normal vector
            if (self.hem=='lh' and mean_normal[0] < 0) or (self.hem=='rh' and mean_normal[0] > 0):
                direction = [mean_normal[0],mean_normal[1],mean_normal[2]]
            else:
                direction = [-1.0*mean_normal[0],-1.0*mean_normal[1],-1.0*mean_normal[2]]

            # project the electrodes to the convex hull of the pial surface
            print('Projection direction vector: ', direction)
        else:
            proj_direction = raw_input('Enter a custom projection direction as a string (lh,rh,top,bottom,front,back,or custom): If none provided, will default to hemisphere: \n')
            if proj_direction == 'custom':
                x = raw_input('Enter projection vector\'s x-component: \n')
                y = raw_input('Enter projection vector\'s y-component: \n')
                z = raw_input('Enter projection vector\'s z-component: \n')
                direction = [float(x),float(y),float(z)]
            elif proj_direction not in ['lh','rh','top','bottom','front','back','custom']:
                direction = self.hem

        #if the grid is placed on the OFC, should create an ROI mesh with only frontal areas 
        roi = ''
        if elecfile_prefix == 'OFC_grid':
            pial_file = os.path.join(self.mesh_dir, self.subj+'_'+self.hem+'_pial.mat')
            orig_elec_file = os.path.join(self.elecs_dir, 'individual_elecs', elecfile_prefix+'_orig.mat')
            surface_warp_scripts_dir = os.path.join(self.img_pipe_dir, 'surface_warping_scripts')
            gyri_labels_dir = os.path.join(self.subj_dir, self.subj, 'label', 'gyri')
            if not os.path.isdir(gyri_labels_dir):
                os.mkdir(gyri_labels_dir)
                # This version of mri_annotation2label uses the coarse labels from the Desikan-Killiany Atlas
                os.system('mri_annotation2label --subject %s --hemi %s --surface pial --outdir %s'\
                    %(self.subj, self.hem, gyri_labels_dir))
            create_roi = matlab.MatlabCommand()
            create_roi.inputs.script = "addpath(genpath('%s'));\
                                        load('%s'); load('%s'); \
                                        make_ofc_roi('%s', '%s',cortex,0); "\
                                        %(surface_warp_scripts_dir, orig_elec_file, \
                                          pial_file, self.subj, self.hem)
            create_roi.run()
            roi = 'ofc_'

        print('::: Loading Mesh data :::')
        print os.path.join(self.subj_dir, self.subj, Meshes, '%s_%s_trivert.mat'%(self.hem, surf_type))
        cortex = scipy.io.loadmat(os.path.join(self.subj_dir, self.subj, Meshes, '%s_%s_trivert.mat'%(self.hem, surf_type)))
        tri, vert = cortex['tri'], cortex['vert']

        print('::: Projecting electrodes to mesh :::')
        elecmatrix = scipy.io.loadmat(os.path.join(self.subj_dir, self.subj, 'elecs', 'individual_elecs', '%s_orig.mat'%(elecfile_prefix)))['elecmatrix']
        print direction
        elecs_proj = project_electrodes_anydirection(tri, vert, elecmatrix, direction)
        scipy.io.savemat(os.path.join(self.subj_dir, self.subj, 'elecs', 'individual_elecs', '%s.mat'%(elecfile_prefix)),{'elecmatrix':elecs_proj})
        print('::: Done :::')

        #move files to preproc subfolder
        if grid:
            corner_file = os.path.join(self.elecs_dir, 'individual_elecs', elecfile_prefix+'_corners.mat')
            orig_file = os.path.join(self.elecs_dir, 'individual_elecs', elecfile_prefix+'_orig.mat')
            preproc_dir = os.path.join(self.elecs_dir, 'individual_elecs', 'preproc') 
            print('Moving %s to %s'%(orig_file, preproc_dir))
            os.system('mv %s %s'%(corner_file, preproc_dir))
            os.system('mv %s %s'%(orig_file, preproc_dir))
        return 

    def get_clinical_grid(self):
        '''Loads in HD grid coordinates and use a list
        of indices to extract downsampled grid.'''

        # load in clinical grid indices
        clingrid = scipy.io.loadmat(os.path.join(self.img_pipe_dir, 'SupplementalFiles','clingrid_inds.mat'))
        clingrid = clingrid.get('inds')

        # load in hd grid coordinates
        hd = scipy.io.loadmat(os.path.join(self.elecs_dir,'hd_grid.mat'))['elecmatrix']

        # get clinical grid coordinates
        clinicalgrid = hd[clingrid]
        clinicalgrid = clinicalgrid[0,:,:]

        # save new coordinates
        scipy.io.savemat(os.path.join(self.elecs_dir, 'clinical_grid.mat'), {'elecmatrix': clinicalgrid})

        return clinicalgrid


    def get_subcort(self):
        '''Obtains .mat files for vertex and triangle
           coords of all subcortical freesurfer segmented meshes'''

        # set ascii dir name
        subjAscii_dir = os.path.join(self.subj_dir, self.subj, 'ascii')
        if not os.path.isdir(subjAscii_dir):
            os.mkdir(subjAscii_dir)

        # tessellate all subjects freesurfer subcortical segmentations
        print('::: Tesselating freesurfer subcortical segmentations from aseg using aseg2srf... :::')
        print(os.path.join(self.img_pipe_dir, 'SupplementalScripts', 'aseg2srf.sh'))
        os.system(os.path.join(self.img_pipe_dir, 'SupplementalScripts', 'aseg2srf.sh') + ' -s "%s" -l "4 5 10 11 12 13 17 18 26 \
                 28 43 44  49 50 51 52 53 54 58 60 14 15 16" -d' % (self.subj))

        # get list of all .srf files and change fname to .asc
        srf_list = list(set([fname for fname in os.listdir(subjAscii_dir)]))
        asc_list = list(set([fname.replace('.srf', '.asc') for fname in srf_list]))
        asc_list.sort()
        for fname in srf_list:
            new_fname = fname.replace('.srf', '.asc')
            os.system('mv %s %s'%(os.path.join(subjAscii_dir,fname), os.path.join(subjAscii_dir,new_fname)))

        # convert all ascii subcortical meshes to matlab vert, tri coords
        subcort_list = ['aseg_058.asc', 'aseg_054.asc', 'aseg_050.asc',
                        'aseg_052.asc', 'aseg_053.asc', 'aseg_051.asc', 'aseg_049.asc',
                        'aseg_043.asc', 'aseg_044.asc', 'aseg_060.asc', 'aseg_004.asc',
                        'aseg_005.asc', 'aseg_010.asc', 'aseg_011.asc', 'aseg_012.asc',
                        'aseg_013.asc', 'aseg_017.asc', 'aseg_018.asc', 'aseg_026.asc',
                        'aseg_028.asc', 'aseg_014.asc', 'aseg_015.asc', 'aseg_016.asc']

        nuc_list = ['rAcumb', 'rAmgd', 'rCaud', 'rGP', 'rHipp', 'rPut', 'rThal',
                    'rLatVent', 'rInfLatVent', 'rVentDienceph', 'lLatVent', 'lInfLatVent',
                    'lThal', 'lCaud', 'lPut',  'lGP', 'lHipp', 'lAmgd', 'lAcumb', 'lVentDienceph',
                    'lThirdVent', 'lFourthVent', 'lBrainStem']

        subcort_dir = os.path.join(self.mesh_dir,'subcortical')     
        if not os.path.isdir(subcort_dir):      
            print('Creating directory %s'%(subcort_dir))        
            os.mkdir(subcort_dir)

        print('::: Converting all ascii segmentations to matlab tri-vert :::')
        for i in range(len(subcort_list)):
            subcort = os.path.join(subjAscii_dir, subcort_list[i])
            nuc = os.path.join(subcort_dir, nuc_list[i])
            self.subcortFs2mlab(subcort, nuc)

    def subcortFs2mlab(self, subcort, nuc):
        '''Function to convert freesurfer ascii subcort segmentations
           to triangular mesh array .mat style.'''

        # use freesurfer mris_convert to get ascii subcortical surface
        subcort_ascii = subcort

        # clean up ascii file and extract matrix dimensions from header
        subcort = open(subcort_ascii, 'r')
        subcort_mat = subcort.readlines()
        subcort.close()
        subcort_mat.pop(0)  # get rid of comments in header
        # get rid of new line char
        subcort_mat = [item.strip('\n') for item in subcort_mat]

        # extract inds for vert and tri
        subcort_inds = subcort_mat.pop(0)
        # seperate inds into two strings
        subcort_inds = subcort_inds.split(' ')
        subcort_inds = [int(i) for i in subcort_inds]  # convert string to ints

        # get rows for vertices only, strip 0 column, and split into seperate
        # strings
        subcort_vert = [item.strip(' 0')
                        for item in subcort_mat[:subcort_inds[0]]]
        subcort_vert = [item.split('  ')
                        for item in subcort_vert]  # seperate strings

        # create containers for each column in vert matrix
        x = []
        y = []
        z = []

        # fill the containers with float values of the strings in each column
        for i in subcort_vert:
            x.append(float(i[0]))
            y.append(float(i[1]))
            z.append(float(i[2]))

        # convert to scipy mat
        x = scipy.mat(x)
        y = scipy.mat(y)
        z = scipy.mat(z)

        # concat columns to one n x 3 matrix
        x = x.transpose()
        y = y.transpose()
        z = z.transpose()
        subcort_vert = scipy.column_stack((x, y, z))
        #scipy.io.savemat('%s_subcort_vert.mat' % (nuc), {'vert': subcort_vert})  # save vert matrix

        # get rows for triangles only, strip 0 column, and split into seperate
        # strings
        subcort_tri = [item[:-2] for item in subcort_mat[subcort_inds[0] + 1:]]
        subcort_tri = [item.split(' ')
                       for item in subcort_tri]  # seperate strings

        # create containers for each column in vert matrix
        x = []
        y = []
        z = []

        # fill the containers with float values of the strings in each column
        for i in subcort_tri:
            x.append(int(i[0]))
            y.append(int(i[1]))
            z.append(int(i[2]))

        # convert to scipy mat
        x = scipy.mat(x)
        y = scipy.mat(y)
        z = scipy.mat(z)

        # concat columns to one n x 3 matrix
        x = x.transpose()
        y = y.transpose()
        z = z.transpose()
        subcort_tri = scipy.column_stack((x, y, z))
        #subcort_tri = subcort_tri 
        outfile = '%s_subcort_trivert.mat' % (nuc)
        scipy.io.savemat(outfile, {'tri': subcort_tri, 'vert': subcort_vert})  # save tri/vert matrix

        # convert inds to scipy mat
        subcort_inds = scipy.mat(subcort_inds)
        scipy.io.savemat('%s_subcort_inds.mat' %
                         (nuc), {'inds': subcort_inds})  # save inds
        
        out_file_struct = '%s_subcort.mat' % (nuc)
        
        cortex = {'tri': subcort_tri+1, 'vert': subcort_vert}
        scipy.io.savemat(out_file_struct, {'cortex': cortex})

    def make_elecs_all(self):
        '''Interactively creates a .mat file with the montage and coordinates of 
        all the elecs files in the /elecs_individual folder.
        '''
        done = False
        short_names,long_names, elec_types, elecmatrix_all = [],[],[], []
        while done == False:
            num_empty_rows = raw_input('Are you adding a row that will be NaN in the elecmatrix? If not, press enter. If so, enter the number of empty rows to add: \n')
            if len(num_empty_rows):
                num_empty_rows = int(num_empty_rows)
                short_name_prefix = raw_input('What is the short name prefix?\n')
                short_names.extend([short_name_prefix for i in range(1,num_empty_rows+1)])
                long_name_prefix = raw_input('What is the long name prefix?\n')
                long_names.extend([long_name_prefix for i in range(1,num_empty_rows+1)])
                elec_type = raw_input('What is the type of the device?\n')
                elec_types.extend([elec_type for i in range(num_empty_rows)])
                elecmatrix_all.append(np.ones((num_empty_rows,3))*np.nan)
            else:
                short_name_prefix = raw_input('What is the short name prefix of the device?\n')
                long_name_prefix = raw_input('What is the long name prefix of the device?\n')
                elec_type = raw_input('What is the type of the device?\n')     
                try:
                    file_name = raw_input('What is the filename of the device\'s electrode coordinate matrix?\n')
                    indiv_file = os.path.join(self.elecs_dir,'individual_elecs', file_name)
                    elecmatrix = scipy.io.loadmat(indiv_file)['elecmatrix']
                except IOError:
                    file_name = raw_input('Sorry, that file was not found. Please enter the correct filename of the device: \n ')
                    indiv_file = os.path.join(self.elecs_dir,'individual_elecs', file_name)
                    elecmatrix = scipy.io.loadmat(indiv_file)['elecmatrix']
                num_elecs = elecmatrix.shape[0]
                elecmatrix_all.append(elecmatrix)
                short_names.extend([short_name_prefix+str(i) for i in range(1,num_elecs+1)])
                long_names.extend([long_name_prefix+str(i) for i in range(1,num_elecs+1)])
                elec_types.extend([elec_type for i in range(num_elecs)])
            completed = raw_input('Finished entering devices? Enter \'y\' if finished.')
            if completed=='y':
                done = True 
        outfile = raw_input('What filename would you like to save out to?\n')
        elecmatrix_all = np.vstack(elecmatrix_all)
        eleclabels = np.ones(elecmatrix_all.shape,dtype=np.object)
        eleclabels[:,0] = short_names
        eleclabels[:,1] = long_names
        eleclabels[:,2] = elec_types
        label_outfile = os.path.join(self.elecs_dir, '%s.mat'%(outfile))
        scipy.io.savemat(label_outfile,{'eleclabels':eleclabels,'elecmatrix':elecmatrix_all})

    def edit_elecs_all(self, revision_dict, elecfile_prefix='TDT_elecs_all'):
        '''Edit the anatomy matrix of the elecfile_prefix. 
        In each entry of the revision_dict, the key is the anatomical label you'd like to impose on the value, which is a list of 0-indexed electrode numbers.
        For example, edit_elecs_all({'superiortemporal':[3,4,5],'precentral':[23,25,36]}).
        '''
        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        elecs_all = scipy.io.loadmat(elecfile)
        anatomy = elecs_all['anatomy']
        for k,v in revision_dict.items():
            if self.zero_indexed_electrodes is False:
                v = np.array(v)+1
            for elec_num in v:
                anatomy[elec_num,3] = k
        elecs_all['anatomy'] = anatomy
        scipy.io.savemat(elecfile, elecs_all)

    def nearest_electrode_vert(self, cortex_verts, elecmatrix):
        ''' Find the vertex on a mesh that is closest to the given electrode
        coordinates.'''

        nchans = elecmatrix.shape[0]
        d = np.zeros((nchans, cortex_verts.shape[0]))

        # Find the distance between each electrode and all possible vertices
        # on the surface mesh
        for chan in np.arange(nchans):
            d[chan,:] = np.sqrt((elecmatrix[chan,0] - cortex_verts[:,0])**2 + \
                                (elecmatrix[chan,1] - cortex_verts[:,1])**2 + \
                                (elecmatrix[chan,2] - cortex_verts[:,2])**2)

        # Find the index of the vertex nearest to each electrode
        vert_inds = np.argmin(d, axis = 1)
        nearest_verts = cortex_verts[vert_inds,:]

        return vert_inds, nearest_verts

    def label_elecs(self, elecfile_prefix='TDT_elecs_all', atlas_surf='desikan-killiany', atlas_depth='destrieux'):
        ''' Automatically labels electrodes based on the freesurfer annotation file.
        Assumes TDT_elecs_all.mat or clinical_elecs_all.mat files
        Uses both the Desikan-Killiany Atlas and the Destrieux Atlas, as described 
        here: https://surfer.nmr.mgh.harvard.edu/fswiki/CorticalParcellation'''

        if atlas_surf == 'desikan-killiany':
            surf_atlas_flag = ''
        elif atlas_surf == 'destrieux':
            surf_atlas_flag = '--a2009s'
        else:
            surf_atlas_flag = ''

        print(self.subj_dir)
        print('Creating labels from the freesurfer annotation file for use in automated electrode labeling')
        gyri_labels_dir = os.path.join(self.subj_dir, self.subj, 'label', 'gyri')
        if not os.path.isdir(gyri_labels_dir):
            os.mkdir(gyri_labels_dir)
         
        # This version of mri_annotation2label uses the coarse labels from the Desikan-Killiany Atlas, unless
        # atlas_surf is 'destrieux', in which case the more detailed labels are used
        os.system('mri_annotation2label --subject %s --hemi %s --surface pial %s --outdir %s'%(self.subj, self.hem, surf_atlas_flag, gyri_labels_dir))
        print('Loading electrode matrix')
        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        elecmatrix = scipy.io.loadmat(elecfile)['elecmatrix']
        
        # Initialize empty variable for indices of grid and strip electrodes
        isnotdepth = []
        
        # Choose only the surface or grid electrodes (if not using hd_grid.mat)
        if elecfile_prefix == 'TDT_elecs_all' or elecfile_prefix == 'clinical_elecs_all':
            elecmontage = scipy.io.loadmat(elecfile)['eleclabels']
            # Make the cell array into something more usable by python
            short_label = []
            long_label = []
            grid_or_depth = []
 
            for r in elecmontage:
                short_label.append(r[0][0]) # This is the shortened electrode montage label
                long_label.append(r[1][0]) # This is the long form electrode montage label
                grid_or_depth.append(r[2][0]) # This is the label for grid, depth, or strip
            
            # These are the indices that won't be used for labeling
            #dont_label = ['EOG','ECG','ROC','LOC','EEG','EKG','NaN','EMG','scalpEEG']
            indices = [i for i, x in enumerate(long_label) if ('EOG' in x or 'ECG' in x or 'ROC' in x or 'LOC' in x or 'EEG' in x or 'EKG' in x or 'NaN' in x or 'EMG' in x or x==np.nan or 'scalpEEG' in x)]
            indices.extend([i for i, x in enumerate(short_label) if ('EOG' in x or 'ECG' in x or 'ROC' in x or 'LOC' in x or 'EEG' in x or 'EKG' in x or 'NaN' in x or 'EMG' in x or x==np.nan or 'scalpEEG' in x)])
            indices.extend([i for i, x in enumerate(grid_or_depth) if ('EOG' in x or 'ECG' in x or 'ROC' in x or 'LOC' in x or 'EEG' in x or 'EKG' in x or 'NaN' in x or 'EMG' in x or x==np.nan or 'scalpEEG' in x)])
            indices.extend(np.where(np.isnan(elecmatrix)==True)[0]) 
            indices = list(set(indices))
            indices_to_use = list(set(range(len(long_label))) - set(indices))

            # Initialize the cell array that we'll store electrode labels in later
            elec_labels_orig = np.empty((len(long_label),4),dtype=np.object)
            elec_labels_orig[:,0] = short_label
            elec_labels_orig[:,1] = long_label
            elec_labels_orig[:,2] = grid_or_depth 
            elec_labels = np.empty((len(indices_to_use),4), dtype = np.object)
            elecmatrix_orig = elecmatrix
            elecmatrix = elecmatrix[indices_to_use,:]
            
            short_label_orig,long_label_orig,grid_or_depth_orig = short_label,long_label,grid_or_depth
            short_label = [i for j, i in enumerate(short_label) if j not in indices]
            long_label = [i for j, i in enumerate(long_label) if j not in indices]
            grid_or_depth = [i for j, i in enumerate(grid_or_depth) if j not in indices]
            elec_labels[:,0] = short_label
            elec_labels[:,1] = long_label
            elec_labels[:,2] = grid_or_depth
            
            # Find the non depth electrodes
            isnotdepth = np.array([r!='depth' for r in grid_or_depth])
            
        # Use the surface label files to get which label goes with each surface vertex
        label_files = glob.glob(os.path.join(gyri_labels_dir, '%s.*.label'%(self.hem)))
        vert_label = {}
        for label in label_files:
            label_name = label.split('.')[1]
            print('Loading label %s'%label_name)
            fid = open(label,'r')
            d = np.genfromtxt(fid, delimiter=' ', \
                              skip_header=2)
            vertnum, x, y, z, junk=d[~np.isnan(d)].reshape((-1,5)).T
            for v in vertnum:
                vert_label[np.int(v)] = label_name.strip()
            fid.close()

        trivert_file = os.path.join(self.mesh_dir, '%s_pial_trivert.mat'%(self.hem))
        cortex_verts = scipy.io.loadmat(trivert_file)['vert']

        # Only use electrodes that are grid or strips
        if len(isnotdepth)>0:
            elecmatrix_new = elecmatrix[isnotdepth,:]
        else:
            elecmatrix_new = elecmatrix

        print('Finding nearest mesh vertex for each electrode')
        vert_inds, nearest_verts = self.nearest_electrode_vert(cortex_verts, elecmatrix_new)

        ## Now make a dictionary of the label for each electrode
        elec_labels_notdepth=[]
        for v in range(len(vert_inds)):
            if vert_inds[v] in vert_label:
                elec_labels_notdepth.append(vert_label[vert_inds[v]].strip())
            else:
                elec_labels_notdepth.append('Unknown')

        if elecfile_prefix == 'TDT_elecs_all' or elecfile_prefix == 'clinical_elecs_all':
            elec_labels[isnotdepth,3] = elec_labels_notdepth
            elec_labels[np.invert(isnotdepth),3] = '' # Set these to an empty string instead of None type
        else:
            elec_labels = np.array(elec_labels_notdepth, dtype = np.object)
        print('Saving electrode labels for surface electrodes to %s'%(elecfile_prefix))
        ## added by BKD so that elec_mat_grid='hd_grid' works. It does not contain elecmontage
        save_dict = {'elecmatrix': elecmatrix, 'anatomy': elec_labels}
        if 'elecmontage' in locals():
            save_dict['eleclabels'] = elecmontage
        else:
            print('electmontage does not exist')
        #scipy.io.savemat('%s/%s/elecs/%s'%(self.subj_dir, self.subj, elecfile_prefix), save_dict)

        if np.any(np.invert(isnotdepth)): # If there are depth electrodes, run this part
            print('*************************************************')
            print('Now doing the depth electrodes')

            # Get the volume corresponding to the labels from the Destrieux atlas, which is more 
            # detailed than Desikan-Killiany (https://surfer.nmr.mgh.harvard.edu/fswiki/CorticalParcellation)
            if atlas_depth == 'desikan-killiany':
                depth_atlas_nm = ''
            elif atlas_depth == 'destrieux':
                depth_atlas_nm = '.a2009s'
            else:
                depth_atlas_nm = '.a2009s'

            aseg_file = os.path.join(self.subj_dir, self.subj, 'mri', 'aparc%s+aseg.mgz'%(depth_atlas_nm))
            dat = nib.freesurfer.load(aseg_file)
            aparc_dat = dat.get_data()
             
            # Define the affine transform to go from surface coordinates to volume coordinates (as CRS, which is
            # the slice *number* as x,y,z in the 3D volume. That is, if there are 256 x 256 x 256 voxels, the
            # CRS coordinate will go from 0 to 255.)
            affine = np.array([[  -1.,    0.,    0.,  128.],
                               [   0.,    0.,    1., -128.],
                               [   0.,   -1.,    0.,  128.],
                               [   0.,    0.,    0.,    1.]])

            elecs_depths = elecmatrix[np.invert(isnotdepth),:]
            intercept = np.ones(len(elecs_depths))
            elecs_ones = np.column_stack((elecs_depths,intercept))

            # Find voxel CRS
            VoxCRS = np.dot(np.linalg.inv(affine), elecs_ones.transpose()).transpose().astype(int)
            # Make meshgrid the same size as aparc_dat (only for gaussian blob version), ignore
            #xx, yy, zz = np.mgrid[0:aparc_dat.shape[0], 0:aparc_dat.shape[1], 0:aparc_dat.shape[2]]
            #unique_labels = np.unique(aparc_dat)
            #unique_labels = unique_labels[unique_labels>0]

            # Get the names of these labels using Freesurfer's lookup table (LUT)
            print("Loading lookup table for freesurfer labels")
            fid = open(os.path.join(self.fs_dir,'FreeSurferColorLUT.txt'))
            LUT = fid.readlines()
            fid.close()

            # Make dictionary of labels
            LUT = [row.split() for row in LUT]
            lab = {}
            for row in LUT:
                if len(row)>1 and row[0][0] is not '#' and row[0][0] is not '\\': # Get rid of the comments
                    lname = row[1]
                    lab[np.int(row[0])] = lname

            # Label the electrodes according to the aseg volume
            nchans = VoxCRS.shape[0]
            anatomy = np.empty((nchans,), dtype=np.object)
            print("Labeling electrodes...")

            for elec in np.arange(nchans):
                anatomy[elec] = lab[aparc_dat[VoxCRS[elec,0], VoxCRS[elec,1], VoxCRS[elec,2]]]
                print("E%d, Vox CRS: [%d, %d, %d], Label #%d = %s"%(elec, VoxCRS[elec,0], VoxCRS[elec,1], VoxCRS[elec,2], aparc_dat[VoxCRS[elec,0], VoxCRS[elec,1], VoxCRS[elec,2]], anatomy[elec]))

            elec_labels[np.invert(isnotdepth),3] = anatomy
            
            #make some corrections b/c of NaNs in elecmatrix
        elec_labels_orig[:,3] = ''
        elec_labels_orig[indices_to_use,3] = elec_labels[:,3] 
        
        print('Saving electrode labels to %s'%(elecfile_prefix))
        scipy.io.savemat(os.path.join(self.elecs_dir, elecfile_prefix+'.mat'), {'elecmatrix': elecmatrix_orig, 'anatomy': elec_labels_orig, 'eleclabels': elecmontage})

        return elec_labels

    def warp_all(self, elecfile_prefix='TDT_elecs_all', warp_depths=True, warp_surface=True, template='cvs_avg35_inMNI152'):
        ''' Warps surface and depth electrodes and runs quality checking functions for them. 
        elecfile_prefix: the name of the .mat file with the electrode coordinates in elecmatrix
        warp_depths: whether to warp depth electrodes
        warp_surface: whether to warp surface electrodes 
        template: which atlas brain to use 
        '''

        print("Using %s as the template for warps"%(template))

        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        elecfile_warped = os.path.join(self.elecs_dir, '%s_warped.mat'%(elecfile_prefix))
        elecfile_nearest_warped = os.path.join(self.elecs_dir, '%s_nearest_warped.mat'%(elecfile_prefix))
        
        if os.path.isfile(elecfile_warped):
            print("The electrodes in %s have already been warped and are in %s"%(elecfile, elecfile_warped))
            return
        
        orig_elecs = scipy.io.loadmat(elecfile)

        if 'depth' in orig_elecs['anatomy'][:,2] and warp_depths:
            #if mri_cvs_register already run, don't run again
            if not os.path.isfile(os.path.join(self.subj_dir, self.subj, 'cvs', 'combined_to'+template+'_elreg_afteraseg-norm.tm3d')):
                self.get_cvsWarp(template)
            else:
                print('%s registration file already created, proceeding to apply the depth warp'%(os.path.join(self.subj_dir, self.subj, 'cvs', 'combined_to'+template+'_elreg_afteraseg-norm.tm3d')))
            if not os.path.isfile(elecfile_nearest_warped):
                self.apply_cvsWarp(elecfile_prefix,template)
            else:
                print "Depth warping has already been applied to the depth electrodes of %s and are in %s"\
                    %(elecfile, elecfile_nearest_warped)
            elecfile_nearest_warped_text = os.path.join(self.elecs_dir, elecfile_prefix+'_nearest_warped.txt')
            elecfile_RAS_text = os.path.join(self.elecs_dir, elecfile_prefix+'_RAS.txt')
            depth_warps = scipy.io.loadmat(elecfile_nearest_warped)
            depth_indices = np.where(orig_elecs['anatomy'][:,2]=='depth')[0]
            orig_elecs['elecmatrix'][depth_indices] = depth_warps['elecmatrix']
            self.check_depth_warps(elecfile_prefix,template)
        
        if warp_surface:
            #if surface warp already run, don't run again
            if not os.path.isfile(os.path.join(self.subj_dir,self.subj,'elecs', elecfile_prefix + '_surface_warped.mat')):
                self.get_surface_warp(elecfile_prefix,template)
            else:
                print('Found %s, not running surface warp again'%(os.path.join(self.subj_dir,self.subj,'elecs', elecfile_prefix + '_surface_warped.mat')))
            elecfile_surface_warped = os.path.join(self.elecs_dir, elecfile_prefix+'_surface_warped.mat')
            surface_warps = scipy.io.loadmat(elecfile_surface_warped)
            surface_indices = np.where(orig_elecs['anatomy'][:,2]!='depth')[0]
            orig_elecs['elecmatrix'][surface_indices] = surface_warps['elecmatrix']

        #if both depth and surface warping have been done, create the combined warp .mat file
        if warp_depths and warp_surface:
            scipy.io.savemat(elecfile_warped,{'elecmatrix':orig_elecs['elecmatrix'],'anatomy':orig_elecs['anatomy']})

            self.plot_recon_anatomy_compare_warp()

            if not os.path.isdir(os.path.join(self.elecs_dir, 'warps_preproc')):
                print('Making preproc directory')
                os.mkdir(os.path.join(self.elecs_dir, 'warps_preproc'))
            preproc_dir = os.path.join(self.elecs_dir, 'warps_preproc')
            if os.path.isfile(elecfile_surface_warped):
                os.system('mv %s %s;' %(elecfile_surface_warped, preproc_dir))
            if os.path.isfile(elecfile_nearest_warped):
                os.system('mv %s %s'%(elecfile_nearest_warped, preproc_dir))
            if os.path.isfile(elecfile_nearest_warped_text):
                os.system('mv %s %s'%(elecfile_nearest_warped_text, preproc_dir))
            if os.path.isfile(elecfile_RAS_text):
                os.system('mv %s %s'%(elecfile_RAS_text, preproc_dir))

    def get_cvsWarp(self,template='cvs_avg35_inMNI152'):
        '''Method for obtaining freesurfer mni coords using mri_cvs_normalize'''

        # run cvs register
        orig = self.subj  # orig is mri in fs orig space

        print('::: Computing Non-linear warping from patient native T1 to fs CVS MNI152 :::')
        os.system('mri_cvs_register --mov %s --template %s --nocleanup --openmp 4' % (orig, template))
        print('cvsWarp COMPUTED')

    def apply_cvsWarp(self, elecfile_prefix='TDT_elecs_all',template_brain='cvs_avg35_inMNI152'):
        ''' Apply the CVS warp from mri_cvs_register to the electrode file of your choice. '''

        elecmatrix = np.empty((0, 4), int)

        fsVox2RAS = np.array(
            [[-1., 0., 0., 128.], [0., 0., 1., -128.], [0., -1., 0., 128.], [0., 0., 0., 1.]])

        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        elecs = scipy.io.loadmat(elecfile)
        anatomy = elecs.get('anatomy')
        depth_indices = np.where(anatomy[:,2]=='depth')[0]
        elec = elecs.get('elecmatrix')[depth_indices,:]

        #### WARP ALL ELECS AT ONCE ####
        print(':::: Computing RAS2Vox and saving %s txt file ::::'%elecfile_prefix)
        if len(elec.shape) > 1:
           intercept = np.ones(len(elec));
           elec = np.column_stack((elec,intercept));
        else:
           elec = np.concatenate((elec, np.ones(1)), axis=0)

        # Convert surface RAS to voxel CRS
        VoxCRS = np.dot(np.linalg.inv(fsVox2RAS), elec.transpose()).transpose()
        orig_file = os.path.join(self.mri_dir, 'orig.mgz')
        affine_file = os.path.join(self.mri_dir, 'affine_subj.txt')
        os.system('mri_info --vox2ras %s > %s'%(orig_file, affine_file))
        affine_subj = np.loadtxt(affine_file)

        affine_template = np.loadtxt(os.path.join(self.subj_dir, '%s_affine_subj.txt'%(template_brain)))

        elec = VoxCRS
        elec = elec[:,0:3]
        nan_elecs = np.isnan(elec)
        elec = np.nan_to_num(elec)

        RAS_text_file = os.path.join(self.elecs_dir, elecfile_prefix+'_RAS.txt')
        np.savetxt(RAS_text_file, elec, fmt='%1.2f\t%1.2f\t%1.2f', delimiter='\t')

        print(':::: Applying Non-linear warping ::::')
        template_brain_file = os.path.join(self.subj_dir, template_brain, 'mri', 'brain.mgz')
        morph_file = os.path.join(self.subj_dir, self.subj, 'cvs', 'combined_to'+template_brain+'_elreg_afteraseg-norm.tm3d')
        nearest_warped_file = os.path.join(self.elecs_dir, elecfile_prefix+'_nearest_warped.txt')
        os.system('applyMorph --template %s --transform %s tract_point_list %s %s nearest' % (template_brain_file,
                                                                                              morph_file,
                                                                                              RAS_text_file,
                                                                                              nearest_warped_file))

        print(':::: Computing Vox2RAS and saving %s_warped mat file ::::'% (elecfile_prefix))
        elec = np.loadtxt(nearest_warped_file)

        elec = np.concatenate((elec, np.ones((elec.shape[0], 1))), axis = 1)

        elecmatrix = np.dot(fsVox2RAS, elec.transpose()).transpose()  

        elecmatrix = elecmatrix[:, 0:3]
        
        # Set the electrodes back to NaN where applicable
        elecmatrix[np.where(nan_elecs[:,0]),:] = np.nan

         # This is for deleting the duplicate row that applyMorph produces for some reason
        if (elecmatrix[-1,:] == elecmatrix[-2,:]).all():
            elecmatrix = elecmatrix[:-1,:]
        nearest_warped_matfile = os.path.join(self.elecs_dir, elecfile_prefix+'_nearest_warped.mat')
        scipy.io.savemat(nearest_warped_matfile, {'elecmatrix': elecmatrix, 'anatomy': anatomy})

    # Method to perform surface warps
    def get_surface_warp(self, basename='TDT_elecs_all', template='cvs_avg35_inMNI152'):
        ''' Perform surface warps on [basename].mat file '''               
        
        elecfile = os.path.join(self.elecs_dir,'%s_surface_warped.mat'%(basename))

        if os.path.isfile(elecfile):
            print("Surface warp file exists")
        else:
            print("Computing surface warp")
            cortex_src = scipy.io.loadmat(self.pial_surf_file[self.hem])
            atlas_file = os.path.join(self.subj_dir, template, 'Meshes', self.hem + '_pial_trivert.mat')
            if not os.path.isfile(atlas_file):
                atlas_patient = freeCoG(subj = template, subj_dir = self.subj_dir, hem = self.hem)
                print("Creating mesh %s"%(atlas_file))
                atlas_patient.convert_fsmesh2mlab()

            cortex_targ = scipy.io.loadmat(atlas_file)
            elecmatrix = scipy.io.loadmat(os.path.join(self.elecs_dir, basename+'.mat'))['elecmatrix']
            anatomy = scipy.io.loadmat(os.path.join(self.elecs_dir, basename+'.mat'))['anatomy']

            print("Finding nearest surface vertex for each electrode")
            vert_inds, nearest_verts = self.nearest_electrode_vert(cortex_src['vert'], elecmatrix)
            elecmatrix = nearest_verts

            print('Warping each electrode separately:')
            elecs_warped = []
            for chan in np.arange(elecmatrix.shape[0]):
                # Open label file for writing
                if anatomy[chan,2] != 'depth':
                    labelname_nopath = '%s.%s.chan%03d.label'%(self.hem, basename, chan)
		    labelpath = os.path.join(self.subj_dir, self.subj, 'label', 'labels_to_warp')
		    if not os.path.isdir(labelpath):
                        os.mkdir(labelpath)
                    labelname = os.path.join(labelpath, labelname_nopath)
                    
                    fid = open(labelname,'w')
                    fid.write('%s\n'%(labelname))
                    
                    # Print header of label file
                    fid.write('#!ascii label  , from subject %s vox2ras=TkReg\n1\n'%(self.subj))
                    fid.write('%i %.9f %.9f %.9f 0.0000000'%(vert_inds[chan], elecmatrix[chan,0], \
                                                            elecmatrix[chan,1], elecmatrix[chan,2]))
                    fid.close()

                    print("Warping ch %d"%(chan))
		    warped_labels_dir = os.path.join(self.subj_dir, template, 'label', 'warped_labels')
		    if not os.path.isdir(warped_labels_dir):
		        os.mkdir(warped_labels_dir)
                    trglabel = os.path.join(warped_labels_dir, '%s.to.%s.%s'%(self.subj, template, labelname_nopath))
                    os.system('mri_label2label --srclabel ' + labelname + ' --srcsubject ' + self.subj + \
                              ' --trgsubject ' + template + ' --trglabel ' + trglabel + ' --regmethod surface --hemi ' + self.hem + \
                              ' --trgsurf pial --paint 6 pial --sd ' + self.subj_dir)

                    # Get the electrode coordinate from the label file
                    fid2 = open(trglabel,'r')
                    coord = fid2.readlines()[2].split() # Get the third line
                    fid2.close()

                    elecs_warped.append([np.float(coord[1]),np.float(coord[2]),np.float(coord[3])])
                else:
                    print("Channel %d is a depth electrode, not warping"%(chan))
                    elecs_warped.append([np.nan, np.nan, np.nan])

                #intersect, t, u, v, xcoor = TriangleRayIntersection(elec, [1000, 0, 0], vert1,vert2,vert3, fullReturn=True)
                
            scipy.io.savemat(elecfile, {'elecmatrix': np.array(elecs_warped), 'anatomy': anatomy})

            print("Surface warp for %s complete. Warped coordinates in %s"%(self.subj, elecfile))

    def check_depth_warps(self, elecfile_prefix='TDT_elecs_all',template='cvs_avg35_inMNI152',atlas_depth='destrieux'):
        ''' Function to check whether warping of depths in mri_cvs_register worked properly. 
        Generates a pdf file with one page for each depth electrode, showing that electrode
        in the original surface space as well as in warped CVS space.  '''
        #get all subj elecs
        RAS_file = os.path.join(self.elecs_dir, '%s_RAS.txt'%(elecfile_prefix))
        subj_elecs,subj_elecnums = self.get_depth_elecs(RAS_file,'\n','\t',elecfile_prefix)

        nearest_warped_file = os.path.join(self.elecs_dir, '%s_nearest_warped.txt'%(elecfile_prefix))
        warped_elecs,warped_elecnums = self.get_depth_elecs(nearest_warped_file,'\n',' ',elecfile_prefix)

        #template brain (cvs)
        if atlas_depth == 'desikan-killiany':
            depth_atlas_nm = ''
        elif atlas_depth == 'destrieux':
            depth_atlas_nm = '.a2009s'
        else:
            depth_atlas_nm = '.a2009s'

        #template brain (cvs)
        cvs_img=nib.freesurfer.load(os.path.join(self.subj_dir, template, 'mri','aparc.a2009s+aseg.mgz'))
        cvs_dat=cvs_img.get_data()

        #subj brain 
        subj_img=nib.freesurfer.load(os.path.join(self.mri_dir,'aparc.a2009s+aseg.mgz'))
        subj_dat=subj_img.get_data()

        pdf = PdfPages(os.path.join(self.elecs_dir, 'depthWarpsQC.pdf'))
        for i in range(len(subj_elecnums)): 
            if subj_elecs[i][0] != 0 and subj_elecs[i][0] != 10000:
                self.plot_elec(subj_elecs[i],warped_elecs[i],subj_dat,cvs_dat,subj_elecnums[i],pdf)
        pdf.close()

    def apply_transform(self, elecfile_prefix, reorient_file):
        ''' Apply an affine transform to an electrode file.  
        for example:
             patient.apply_transform(elecfile_prefix = 'TDT_elecs_all', reorient_file = 'T1_reorient')
        assumes transform is located in the subject's acpc directory
        '''
        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        elecs = scipy.io.loadmat(elecfile)
        elec = elecs.get('elecmatrix')
        anatomy = elecs.get('anatomy')
        
        reorient_file =  os.path.join(self.subj_dir, self.subj, 'acpc', reorient_file+'.mat')
        rmat = scipy.io.loadmat(reorient_file)['M']
        elecs_reoriented = nib.affines.apply_affine(rmat, elec)
        
        print("Saving reoriented electrodes")
        reoriented_elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'_reoriented.mat')
        scipy.io.savemat(reoriented_elecfile, {'elecmatrix': elecs_reoriented, 'anatomy': anatomy})
        print("Done.")

     #helper method to check the cvs depth warps:
    def plot_elec(self,orig_coords,warped_coords,subj_dat,cvs_dat,elec_num,pdf):
        ''' helper method to check the cvs depth warps. Each electrode is one page      
        in the resulting PDF.       
        Top row shows electrodes warped to the CVS brain, bottom row shows the electrodes       
        in their original position on the subject's brain. If the anatomical label      
        for the warped location matches that of the original subject brain, it is counted       
        as a "MATCH" and has a title in green, otherwise it is a "MISMATCH" and is      
        marked with a red title.        
        '''

        fs_lut = os.path.join(self.img_pipe_dir, 'SupplementalFiles', 'FreeSurferLUTRGBValues.npy')
        cmap = matplotlib.colors.ListedColormap(np.load()[:cvs_dat.max()+1,:])

        lookupTable = os.path.join(self.img_pipe_dir, 'SupplementalFiles', 'FreeSurferLookupTable')
        lookup_dict = pickle.load(open(lookupTable,'r'))
        fig = plt.figure(figsize=((30,17)))
        nonzero_indices = np.where(cvs_dat>0)
        offset = 35 #this is how much you want to trim the mri by, there is a lot of empty space

        cvs_dat = cvs_dat[offset:-offset,offset:-offset,offset:-offset]
        cvs_vox_CRS = np.array([warped_coords[0]-offset,warped_coords[1]-offset,warped_coords[2]-offset],dtype='int')

        plt.subplot(2,3,1)
        plt.imshow(cvs_dat[cvs_vox_CRS[0],:,:], cmap=cmap)
        plt.plot(cvs_vox_CRS[2],cvs_vox_CRS[1],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()

        plt.subplot(2,3,2)
        plt.imshow(cvs_dat[:,cvs_vox_CRS[1],:].T, cmap=cmap)
        plt.plot(cvs_vox_CRS[0],cvs_vox_CRS[2],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()
        #t=plt.text(10,25,lookup_dict[cvs_dat[cvs_vox_CRS[0],cvs_vox_CRS[1],cvs_vox_CRS[2]]],color='white',size=25)
        plt.title('CVS brain ' + lookup_dict[cvs_dat[cvs_vox_CRS[0],cvs_vox_CRS[1],cvs_vox_CRS[2]]],size=20)

        plt.subplot(2,3,3)
        plt.imshow(cvs_dat[:,:,cvs_vox_CRS[2]].T, cmap=cmap)   
        plt.plot(cvs_vox_CRS[0],cvs_vox_CRS[1],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()

        subj_dat = subj_dat[offset:-offset,offset:-offset,offset:-offset]
        subj_vox_CRS = np.array([orig_coords[0]-offset,orig_coords[1]-offset,orig_coords[2]-offset],dtype='int')

        ax1=plt.subplot(2,3,4).axes
        plt.imshow(subj_dat[subj_vox_CRS[0],:,:], cmap=cmap)
        plt.plot(subj_vox_CRS[2],subj_vox_CRS[1],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()

        ax2=plt.subplot(2,3,5).axes
        plt.imshow(subj_dat[:,subj_vox_CRS[1],:].T, cmap=cmap)
        plt.plot(subj_vox_CRS[0],subj_vox_CRS[2],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()
        #t=plt.text(10,25,lookup_dict[subj_dat[subj_vox_CRS[0],subj_vox_CRS[1],subj_vox_CRS[2]]],color='white',size=25)
        plt.title(self.subj + ' ' +lookup_dict[subj_dat[subj_vox_CRS[0],subj_vox_CRS[1],subj_vox_CRS[2]]],size=20)

        ax3=plt.subplot(2,3,6).axes
        plt.imshow(subj_dat[:,:,subj_vox_CRS[2]].T, cmap=cmap)   
        plt.plot(subj_vox_CRS[0],subj_vox_CRS[1],'r*',markersize=14,color='#FFFFFF')
        plt.axis('tight'); ax = plt.gca(); ax.set_axis_off()

        if cvs_dat[cvs_vox_CRS[0],cvs_vox_CRS[1],cvs_vox_CRS[2]] != subj_dat[subj_vox_CRS[0],subj_vox_CRS[1],subj_vox_CRS[2]]:
            result = 'MISMATCH'
            color = 'r'
        else:
            result = 'Match'
            color = 'g'
        plt.suptitle('elec ' +str(elec_num) + ' '+result, size=30,color=color)
        pdf.savefig(fig)
        plt.close()

    def get_depth_elecs(self, fpath, delim1, delim2, elecfile_prefix):
        ''' Helper method to check cvs depth warps.  This helper method         
        just loads the electrode coordinates of the depth electrodes from the matlab file            
        '''
        f = open(fpath,'r')
        elecs = (f.read().split(delim1))
        elecs = np.array([e for e in elecs if len(e)>1])
        elecfile = os.path.join(self.elecs_dir, elecfile_prefix+'.mat')
        tdt_elec_types = scipy.io.loadmat(elecfile)['anatomy'][:,2][:-1]
        # depth_elecs = elecs[np.where(tdt_elec_types == 'depth')[0]]
        str_to_float = np.vectorize(lambda x: float(x))
        depth_elecs = str_to_float(np.array([s.split(delim2) for s in elecs]))
        return depth_elecs,np.where(tdt_elec_types == 'depth')[0]

    def get_unique_anatomy(self, elecfile_prefix='TDT_elecs_all'):
        ''' Get a list of all possible electrode anatomical labels for this subject '''
        anat = self.get_elecs(elecfile_prefix=elecfile_prefix)['anatomy'][:,3]
        return np.unique([a[0] for a in anat])

    def get_elecs(self, elecfile_prefix='TDT_elecs_all', roi=None):
        '''
        Utility function to get electrode coordinate matrix of all electrodes in a certain anatomical region.
        To get a list of the anatomical labels you can use, call patient.get_unique_anatomy()
        '''

        e = {'elecmatrix': [], 'anatomy': []}
        elecfile = os.path.join(self.elecs_dir,'%s.mat'%(elecfile_prefix))
        if os.path.isfile(elecfile):
            e = scipy.io.loadmat(elecfile)
            if roi != None:
                roi_indices = np.where(e['anatomy'][:,3]==roi)[0]
                elecmatrix = e['elecmatrix'][roi_indices,:]
                anatomy = e['anatomy'][roi_indices,:]
                e = {'elecmatrix': elecmatrix, 'anatomy': anatomy}
        else:
            print("%s does not exist"%(elecfile))

        return e

    class roi:

        def __init__(self, name, color=(0.8,0.8,0.8), opacity=1.0, representation='surface', gaussian=False):
            '''wrapper class for an ROI
            name: name of the ROI
            color: tuple for the ROI's color where each value is between 0.0 and 1.0. 
            opacity: opacity of the mesh, between 0.0 and 1.0
            representation: 'surface' or 'wireframe'
            gaussian: boolean specifying how to represent electrodes and their weights on this mesh, note that setting gaussian to True 
                      means the mesh color cannot be specified by the user'''
                       
            self.name = name
            self.color = color
            self.opacity = opacity
            self.representation = representation
            self.gaussian = gaussian

    def get_rois(self):
        ''' Method to get the list of all available ROI names for plotting '''
        rois = ['rAcumb', 'rAmgd', 'rCaud', 'rGP', 'rHipp', 'rPut', 'rThal',
                    'rLatVent', 'rInfLatVent', 'rVentDienceph', 'lLatVent', 'lInfLatVent',
                    'lThal', 'lCaud', 'lPut',  'lGP', 'lHipp', 'lAmgd', 'lAcumb', 'lVentDienceph',
                    'lThirdVent', 'lFourthVent', 'lBrainStem', 'pial', 'lh_pial', 'rh_pial']
        return rois

    def plot_brain(self, rois=[roi(name='pial', color=(0.8,0.8,0.8), opacity=1.0, representation='surface', gaussian=False)], elecs=[], weights=[], cmap = 'RdBu', showfig=True):
        '''plots multiple meshes on one figure. Defaults to plotting both hemispheres of the pial surface.
        rois: list of roi objects (create an roi object like so:
              hipp_roi = patient.roi(name='lHipp', color=(0.5,0.1,0.8), opacity=1.0, representation='surface', gaussian=True))
        elecs: electrode coordinate matrix
        weights: weight matrix associated with the electrode coordinate matrix

        Example: 
        >>> pial, hipp = patient.roi('pial',(0.6,0.3,0.6),0.1,'wireframe',True), patient.roi('lHipp',(0.5,0.1,0.8),1.0,'surface',True)
        >>> elecs = patient.get_elecs()['elecmatrix']
        >>> patient.plot_brain(rois=[pial,hipp],elecs=elecs,weights=np.random.uniform(0,1,(elecs.shape[0])))
        '''

        import mayavi
        import plotting.ctmr_brain_plot as ctmr_brain_plot
        
        mayavi.mlab.figure(fgcolor=(0, 0, 0), bgcolor=(1, 1, 1), size=(1200,900))

        #if there are any rois with gaussian set to True, don't plot any elecs as points3d, to avoid mixing gaussian representation
        #if needed, simply add the elecs by calling el_add
        any_gaussian = False

        for roi in rois:
            roi_name, color, opacity, representation, gaussian = roi.name, roi.color, roi.opacity, roi.representation, roi.gaussian

            #if color or opacity == None, then use default values 
            if color == None:
                color = (0.8,0.8,0.8)
            if opacity == None:
                opacity = 1.0
            if representation == None:
                representation = 'surface'
            if gaussian == None:
                gaussian = False
            if gaussian == True:
                any_gaussian = True

            #default roi_name of 'pial' plots both hemispheres' pial surfaces
            if roi_name =='pial' or roi_name == 'rh_pial' or roi_name == 'lh_pial':
                #use pial surface of the entire hemisphere
                lh_pial = scipy.io.loadmat(self.pial_surf_file['lh'])
                rh_pial = scipy.io.loadmat(self.pial_surf_file['rh'])
                if gaussian:
                    if roi_name == 'pial' or roi_name == 'lh_pial':
                        mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(lh_pial['tri'], lh_pial['vert'], color=color, opacity=opacity, elecs=elecs, weights=weights, representation=representation, new_fig=False, cmap=cmap)
                    if roi_name == 'pial' or roi_name == 'rh_pial': 
                        mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(rh_pial['tri'], rh_pial['vert'], color=color, opacity=opacity, elecs=elecs, weights=weights, representation=representation, new_fig=False, cmap=cmap)
                else: 
                    if roi_name == 'pial' or roi_name == 'lh_pial': 
                        mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(lh_pial['tri'], lh_pial['vert'], color=color, opacity=opacity, representation=representation, new_fig=False, cmap=cmap)
                    if roi_name == 'pial' or roi_name == 'rh_pial': 
                        mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(rh_pial['tri'], rh_pial['vert'], color=color, opacity=opacity, representation=representation, new_fig=False, cmap=cmap)
                    
            else: 
                subcort_dir = os.path.join(self.mesh_dir,'subcortical')
                if os.path.isdir(subcort_dir) and '%s_subcort_trivert.mat'%(roi_name) in os.listdir(subcort_dir):
                    roi_mesh = scipy.io.loadmat(os.path.join(subcort_dir,'%s_subcort_trivert.mat'%(roi_name)))
                else:
                    roi_mesh = scipy.io.loadmat(os.path.join(self.mesh_dir,'%s_trivert.mat'%(roi_name)))

                if gaussian:
                    mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(roi_mesh['tri'],roi_mesh['vert'],color=(color), opacity=opacity, elecs=elecs, weights=weights, representation=representation, new_fig=False, cmap=cmap)
                else:
                    mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(roi_mesh['tri'],roi_mesh['vert'],color=(color), opacity=opacity, representation=representation, new_fig=False, cmap=cmap)
        if not any_gaussian and elecs!=[]:
            if weights==[]: #if elecmatrix passed in but no weights specified, default to all ones for the electrode color weights
                elec_colors = np.zeros((elecs.shape[0],3))
                elec_colors[:,0] = 1. #defaults to red elecs if no weights specified
            else:
                # Map the weights onto the current colormap
                elec_colors = cm.get_cmap(cmap)(weights)[:,:3] 
            points, mlab = ctmr_brain_plot.el_add(elecs, color = elec_colors)
        else:
            #if no elecs to add as points3D
            points = None

        if self.hem=='lh':
            azimuth=180
        elif self.hem=='rh':
            azimuth=0
        mlab.view(azimuth, elevation=90)

        if showfig:
            mlab.show()
        return mesh, points, mlab

    def plot_recon_anatomy(self, elecfile_prefix='TDT_elecs_all', template=None, interactive=True, screenshot=False, opacity=1.0):
        import mayavi
        import plotting.ctmr_brain_plot as ctmr_brain_plot
        import SupplementalFiles.FS_colorLUT as FS_colorLUT

        if template == None:
            a = scipy.io.loadmat(self.pial_surf_file[self.hem])
        else:
            template_pial_surf_file = os.path.join(self.subj_dir, template, 'Meshes', self.hem+'_pial_trivert.mat')
            a = scipy.io.loadmat(template_pial_surf_file)

        e = self.get_elecs(elecfile_prefix = elecfile_prefix)

        # Plot the pial surface
        mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(a['tri'], a['vert'], color=(0.8, 0.8, 0.8), opacity=opacity)

        # Add the electrodes, colored by anatomical region
        elec_colors = np.zeros((e['elecmatrix'].shape[0], e['elecmatrix'].shape[1]))

        # Import freesurfer color lookup table as a dictionary
        cmap = FS_colorLUT.get_lut()

        # Make a list of electrode numbers
        if self.zero_indexed_electrodes:
            elec_numbers = np.arange(e['elecmatrix'].shape[0])
        else:
            elec_numbers = np.arange(e['elecmatrix'].shape[0])+1
        if self.hem=='lh':
            label_offset=-1.0
        elif self.hem=='rh':
            label_offset=1.0

        # Find all the unique brain areas in this subject
        brain_areas = np.unique(e['anatomy'][:,3])

        # Loop through unique brain areas and find the appropriate color for each brain area from the color LUT dictionary
        for b in brain_areas:
            # Add relevant extra information to the label if needed for the color LUT
            if b != 'NaN':
                this_label = b[0]
                if b[0][0:3]!='ctx' and b[0][0:4] != 'Left' and b[0][0:5] != 'Right' and b[0][0:5] != 'Brain' and b[0] != 'Unknown':
                    this_label = 'ctx-%s-%s'%(self.hem, b[0])
                    print(this_label)
                
                if this_label != '':
                    if this_label not in cmap:
                        #in case the label was manually assigned, and not found in the LUT colormap dictionary
                        el_color = matplotlib.cm.get_cmap('viridis').colors[int(float(np.where(brain_areas==b)[0])/float(len(brain_areas)))]
                    else:
                        el_color = np.array(cmap[this_label])/255.
            elec_indices = np.where(e['anatomy'][:,3]==b)[0]
            elec_colors[elec_indices,:] = el_color

        if self.zero_indexed_electrodes:
            elec_nums = range(e['elecmatrix'].shape[0])
        else:
            elec_nums = range(1,e['elecmatrix'].shape[0]+1)

        ctmr_brain_plot.el_add(e['elecmatrix'],elec_colors,numbers=elec_nums, label_offset=label_offset)

        if self.hem=='lh':
            azimuth=180
        elif self.hem=='rh':
            azimuth=0
        mlab.view(azimuth, elevation=90)

        mlab.title('%s recon anatomy'%(self.subj),size=0.3)

        arr = mlab.screenshot(antialiased=True)
        if screenshot:
            plt.figure(figsize=(20,10))
            plt.imshow(arr, aspect='equal')
            plt.axis('off')
            plt.show()
        if interactive:
            mlab.show()
        else:
            mlab.close()
        return mesh, mlab

    def plot_erps(self, erp_matrix, time_scale_factor=0.03, z_scale_factor=3.0):
        import mayavi
        import plotting.ctmr_brain_plot as ctmr_brain_plot
        import SupplementalFiles.FS_colorLUT as FS_colorLUT
        #use mean normal vector
        #no anat color map option
        #time *-1 if left hemisphere
        #centering the time series more accurately
        #plotting all lines without the loop
        #mean normal as the offset for the erp lines' x-coordinate

        cmap = FS_colorLUT.get_lut()

        anatomy_labels = scipy.io.loadmat(os.path.join(self.elecs_dir,'TDT_elecs_all.mat'))['anatomy'][:,3]

        mesh, points, mlab = self.plot_brain(showfig=False)
        elecmatrix = self.get_elecs()['elecmatrix']
        for c in range(erp_matrix.shape[1]):
            print c
            b = anatomy_labels[c]
            if b != 'NaN':
                this_label = b[0]
                if b[0][0:3]!='ctx' and b[0][0:4] != 'Left' and b[0][0:5] != 'Right' and b[0][0:5] != 'Brain' and b[0] != 'Unknown':
                    this_label = 'ctx-%s-%s'%('lh', b[0])

                if this_label != '':
                    if this_label not in cmap:
                        #in case the label was manually assigned, and not found in the LUT colormap dictionary
                        el_color = matplotlib.cm.get_cmap('viridis').colors[int(float(np.where(brain_areas==b)[0])/float(len(brain_areas)))]
                    else:
                        el_color = np.array(cmap[this_label])/255.

            el_color = tuple(el_color)
            elec_coord = elecmatrix[c,:]
            mlab.points3d(elec_coord[0],elec_coord[1],elec_coord[2], scale_factor = 0.25, color = (1.0, 0.0, 0.0), resolution=25)
            erp = erp_matrix[:,c]
            mlab.plot3d(np.array([elec_coord[0] for i in range(erp_matrix.shape[0])])-2.0, ((np.array([i for i in range(-50,55)]))*time_scale_factor+elec_coord[1])[::-1], erp*z_scale_factor+elec_coord[2],\
                            figure=mlab.gcf(),color=(el_color),tube_radius=0.15,tube_sides=3)

        mlab.show()

    def plot_recon_anatomy_compare_warped(self, template, elecfile_prefix='TDT_elecs_all',interactive=True, screenshot=False, opacity=1.0):
        import mayavi
        import plotting.ctmr_brain_plot as ctmr_brain_plot
        import SupplementalFiles.FS_colorLUT as FS_colorLUT

        subj_brain = scipy.io.loadmat(self.pial_surf_file[self.hem])
        template_pial_surf_file = os.path.join(self.subj_dir, template, 'Meshes', self.hem+'_pial_trivert.mat')
        template_brain = scipy.io.loadmat(template_pial_surf_file)

        # Get native space and warped electrodes
        subj_e = self.get_elecs(elecfile_prefix = elecfile_prefix)
        template_e = self.get_elecs(elecfile_prefix = elecfile_prefix+'_warped')
        
        subj_brain_width = np.abs(np.max(subj_brain['vert'][:,1])-np.min(subj_brain['vert'][:,1]))
        template_brain_width = np.abs(np.max(template_brain['vert'][:,1])-np.min(template_brain['vert'][:,1]))
        
        subj_e['elecmatrix'][:,1] = subj_e['elecmatrix'][:,1]-((subj_brain_width+template_brain_width)/4)
        template_e['elecmatrix'][:,1] = template_e['elecmatrix'][:,1]+((subj_brain_width+template_brain_width)/4)

        subj_brain['vert'][:,1] = subj_brain['vert'][:,1]-((subj_brain_width+template_brain_width)/4)
        template_brain['vert'][:,1] = template_brain['vert'][:,1]+((subj_brain_width+template_brain_width)/4)
        
        # Plot the pial surface
        subj_mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(subj_brain['tri'], subj_brain['vert'], color=(0.8, 0.8, 0.8))
        template_mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(template_brain['tri'], template_brain['vert'], color=(0.8, 0.8, 0.8),new_fig=False)
        
        # Add the electrodes, colored by anatomical region
        elec_colors = np.zeros((subj_e['elecmatrix'].shape[0], subj_e['elecmatrix'].shape[1]))

        # Import freesurfer color lookup table as a dictionary
        cmap = FS_colorLUT.get_lut()

        # Make a list of electrode numbers
        if self.zero_indexed_electrodes:
            elec_numbers = np.arange(subj_e['elecmatrix'].shape[0])
        else:
            elec_numbers = np.arange(subj_e['elecmatrix'].shape[0])+1

        # Find all the unique brain areas in this subject
        brain_areas = np.unique(subj_e['anatomy'][:,3])

        # Loop through unique brain areas and plot the electrodes in each brain area
        for b in brain_areas:
            # Add relevant extra information to the label if needed for the color LUT
            if b != 'NaN':
                this_label = b[0]
                if b[0][0:3]!='ctx' and b[0][0:4] != 'Left' and b[0][0:5] != 'Right' and b[0][0:5] != 'Brain' and b[0] != 'Unknown':
                    this_label = 'ctx-%s-%s'%(self.hem, b[0])
                    print(this_label)
                    if this_label != '':
                        if this_label not in cmap:
                            #in case the label was manually assigned, and not found in the LUT colormap dictionary
                            el_color = matplotlib.cm.get_cmap('viridis').colors[int(float(np.where(brain_areas==b)[0])/float(len(brain_areas)))]
                        else:
                            el_color = np.array(cmap[this_label])/255.
                    ctmr_brain_plot.el_add(np.atleast_2d(subj_e['elecmatrix'][subj_e['anatomy'][:,3]==b,:]), 
                                           color=tuple(el_color), numbers=elec_numbers[subj_e['anatomy'][:,3]==b])

                    ctmr_brain_plot.el_add(np.atleast_2d(template_e['elecmatrix'][subj_e['anatomy'][:,3]==b,:]), 
                                           color=tuple(el_color), numbers=elec_numbers[subj_e['anatomy'][:,3]==b])
        if self.hem=='lh':
            azimuth=180
        elif self.hem=='rh':
            azimuth=0
        mlab.view(azimuth, elevation=90)

        #adjust transparency of brain mesh
        subj_mesh.actor.property.opacity = opacity
        template_mesh.actor.property.opacity = opacity 

        arr = mlab.screenshot(antialiased=True)
        if screenshot:
            plt.figure(figsize=(20,10))
            plt.imshow(arr, aspect='equal')
            plt.axis('off')
            plt.show()
        if interactive:
            mlab.show()
        else:
            mlab.close()
        return subj_mesh, template_mesh, mlab

    


import numpy as np
import scipy.io
def get_elecs_anat(subj_dir,subj,region):
    tdt = scipy.io.loadmat('%s/%s/elecs/TDT_elecs_all.mat'%(subj_dir,subj))
    return tdt['elecmatrix'][np.where(tdt['anatomy'][:,3]==region)[0],:]

def ctmr_plot(subj_dir,subj,hem, elecs, weights=None,interactive=False):
    #need ctmr_brain_plot, mayavi, vtk
    import ctmr_brain_plot
    from mayavi import mlab
    from matplotlib import cm
    import scipy.io
    import numpy as np
    import matplotlib.pyplot as plt
    
    a = scipy.io.loadmat('%s/%s/Meshes/%s_pial_trivert.mat'%(subj_dir, subj,hem));
    if weights==None:
        weights = np.ones((elecs.shape[0]))*-1.
    mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(a['tri'], a['vert'], elecs=elecs,weights=weights,color=(0.8, 0.8, 0.8),cmap='RdBu')

    mesh.actor.property.opacity = 1.0 # Make brain semi-transparent

    # View from the side
    if hem=='lh':
        azimuth=180
    elif hem=='rh':
        azimuth=0
    mlab.view(azimuth, elevation=90)
    arr = mlab.screenshot(antialiased=True)
    plt.figure(figsize=(20,10))
    plt.imshow(arr, aspect='equal')
    plt.axis('off')
    plt.show()
    if interactive:
        mlab.show()
    else:
        mlab.close()
    return mesh,mlab


#!/usr/bin/env python

import matplotlib
from pyface.qt import QtGui, QtCore
matplotlib.use('Qt4Agg') 
from matplotlib import pyplot as plt
plt.rcParams['keymap.save'] = '' # Unbind 's' key saving
plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = ['Helvetica','Verdana','Bitstream Vera Sans','sans-serif']

from matplotlib import cm
import matplotlib.colors as mcolors
import numpy as np
import nibabel as nib
import scipy.ndimage
import sys
#from PyQt4 import QtCore, QtGui
import matplotlib.patches as mpatches
import scipy.io
import os
import warnings

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

warnings.filterwarnings('ignore')

class electrode_picker:
	'''
	electrode_picker.py defines a class [electrode_picker] that allows the 
	user to take a co-registered CT and MRI scan and identify electrodes
	based on sagittal, coronal, and axial views of the scans as well as
	a maximum intensity projection of the CT scan. Inputs are the 
	subject directory and the hemisphere of implantation.  If stereo-EEG,
	choose 'stereo' as the hemisphere

	Usage: 
		python electrode_picker.py '/usr/local/freesurfer/subjects/S1' 'rh'

	This assumes that you have processed your data using freesurfer's pipeline
	and that you have a coregistered MRI and CT in subj_dir (e.g. '/usr/local/freesurfer/subjects/S1')
	as [subj_dir]/mri/brain.mgz and [subj_dir]/CT/rCT.nii

	Written by Liberty Hamilton, 2017

	'''
	def __init__(self, subj_dir, hem):
		'''
		Initialize the electrode picker with the user-defined MRI and co-registered
		CT scan in [subj_dir].  Images will be displayed using orientation information 
		obtained from the image header. Images will be resampled to dimensions 
		[256,256,256] for display.
		We will also listen for keyboard and mouse events so the user can interact
		with each of the subplot panels (zoom/pan) and add/remove electrodes with a 
		keystroke.
		'''
		QtCore.pyqtRemoveInputHook()
		self.subj_dir = subj_dir
		if hem == 'stereo':
			hem = 'lh' # For now, set to lh because hemisphere isn't used in stereo case
		self.hem = hem
		self.img = nib.load(os.path.join(subj_dir, 'mri', 'brain.mgz'))
		self.ct = nib.load(os.path.join(subj_dir, 'CT', 'rCT.nii'))
		pial_fill = os.path.join(subj_dir, 'surf', '%s.pial.filled.mgz'%(self.hem))
		if not os.path.isfile(pial_fill):
			pial_surf = os.path.join(subj_dir, 'surf', '%s.pial'%(self.hem))
			os.system('mris_fill -c -r 1 %s %s'%(pial_surf, pial_fill))
		self.pial_img = nib.load(pial_fill)
		
		# Get affine transform 
		self.affine = self.img.affine
		self.fsVox2RAS = np.array([[-1., 0., 0., 128.], 
								   [0., 0., 1., -128.], 
								   [0., -1., 0., 128.], 
								   [0., 0., 0., 1.]])
		
		# Apply orientation to the MRI so that the order of the dimensions will be
		# sagittal, coronal, axial
		self.codes = nib.orientations.axcodes2ornt(nib.orientations.aff2axcodes(self.affine))
		img_data = nib.orientations.apply_orientation(self.img.get_data(), self.codes)
		self.voxel_sizes = nib.affines.voxel_sizes(self.affine)
		nx,ny,nz = np.array(img_data.shape, dtype='float')

		self.inv_affine = np.linalg.inv(self.affine)
		self.img_clim = np.percentile(img_data, (1., 99.))

		# Apply orientation to pial surface fill
		self.pial_codes = nib.orientations.axcodes2ornt(nib.orientations.aff2axcodes(self.pial_img.affine))
		pial_data = nib.orientations.apply_orientation(self.pial_img.get_data(), self.pial_codes)
		pial_data = scipy.ndimage.binary_closing(pial_data)

		# Apply orientation to the CT so that the order of the dimensions will be
		# sagittal, coronal, axial
		self.ct_codes =nib.orientations.axcodes2ornt(nib.orientations.aff2axcodes(self.ct.affine))
		ct_data = nib.orientations.apply_orientation(self.ct.get_data(), self.ct_codes)

		# Threshold the CT so only bright objects (electrodes) are visible
		ct_data[ct_data < 1000] = np.nan
		cx,cy,cz=np.array(self.ct.shape, dtype='float')
		

		# Resample both images to the highest resolution
		voxsz = (256, 256, 256)
		if self.ct.shape != voxsz:
			print("Resizing voxels in CT")
			ct_data = scipy.ndimage.zoom(ct_data, [voxsz[0]/cx, voxsz[1]/cy, voxsz[2]/cz])
		if self.img.shape != voxsz:
			print("Resizing voxels in MRI")
			img_data = scipy.ndimage.zoom(img_data, [voxsz[0]/nx, voxsz[1]/ny, voxsz[2]/nz])
		
		self.ct_data = ct_data
		self.img_data = img_data
		self.pial_data = pial_data
		self.elec_data = np.nan+np.zeros((img_data.shape))
		self.bin_mat = '' # binary mask for electrodes
		self.device_num = 0 # Start with device 0, increment when we add a new electrode name type
		self.device_name = ''
		self.devices = [] # This will be a list of the devices (grids, strips, depths)
		self.elec_num = dict()
		self.elecmatrix = dict()# This will be the electrode coordinates 
		self.legend_handles = [] # This will hold legend entries

		self.imsz = [256, 256, 256]
		self.ctsz = [256, 256, 256]

		self.current_slice = np.array([self.imsz[0]/2, self.imsz[1]/2, self.imsz[2]/2], dtype=np.float)
		
		self.fig=plt.figure(figsize=(12,10))
		self.fig.canvas.set_window_title('Electrode Picker')
		thismanager = plt.get_current_fig_manager()
		thismanager.window.setWindowIcon(QtGui.QIcon((os.path.join('icons','leftbrain_blackbg.png'))))
		
		self.im = []
		self.ct_im = []
		self.elec_im = []
		self.pial_im = []

		self.cursor = []
		self.cursor2 = []

		im_ranges = [[0, self.imsz[1], 0, self.imsz[2]],
					 [0, self.imsz[0], 0, self.imsz[2]],
					 [0, self.imsz[0], 0, self.imsz[1]]]
		im_labels = [['Inferior','Posterior'],
					 ['Inferior','Left'],
					 ['Posterior','Left']]

		self.ax = []
		self.contour = [False, False, False]
		self.pial_surf_on = True # Whether pial surface is visible or not
		
		# This is the current slice for indexing (as integers so python doesnt complain)
		cs = np.round(self.current_slice).astype(np.int)

		# Plot sagittal, coronal, and axial views 
		for i in np.arange(3):
			self.ax.append(self.fig.add_subplot(2,2,i+1))
			self.ax[i].set_axis_bgcolor('k')
			if i==0:
				imdata = img_data[cs[0],:,:].T
				ctdat  = ct_data[cs[0],:,:].T
				edat   = self.elec_data[cs[0],:,:].T
				pdat   = self.pial_data[cs[0],:,:].T
			elif i==1:
				imdata = img_data[:,cs[1],:].T
				ctdat  = ct_data[:,cs[1],:].T
				edat   = self.elec_data[:,cs[1],:].T
				pdat   = self.pial_data[:,cs[1],:].T
			elif i==2:
				imdata = img_data[:,:,cs[2]].T
				ctdat  = ct_data[:,:,cs[2]].T
				edat   = self.elec_data[:,:,cs[2]].T
				pdat   = self.pial_data[:,:,cs[2]].T

			# Show the MRI data in grayscale
			self.im.append(plt.imshow(imdata, cmap=cm.gray, aspect='auto'))

			# Overlay the CT on top in "hot" colormap, slightly transparent
			self.ct_im.append(plt.imshow(ctdat, cmap=cm.hot, aspect='auto',alpha=0.5, vmin=1000, vmax=3000))
			
			# Overlay the electrodes image on top (starts as NaNs, is eventually filled in)
			self.elec_colors = mcolors.LinearSegmentedColormap.from_list('elec_colors', np.vstack (( cm.Set1(np.linspace(0., 1, 9)), cm.Set2(np.linspace(0., 1, 8)) )) )
			self.elec_im.append(plt.imshow(edat, cmap=self.elec_colors, aspect='auto', alpha=1, vmin=0, vmax=17))
			
			# Overlay the pial surface
			self.pial_im.append(self.ax[i].contour(pdat, linewidths=0.5, colors = 'y'))
			self.contour[i] = True

			# Plot a green cursor
			self.cursor.append(plt.plot([cs[1], cs[1]], [self.ax[i].get_ylim()[0]+1, self.ax[i].get_ylim()[1]-1], color=[0, 1, 0] ))
			self.cursor2.append(plt.plot([self.ax[i].get_xlim()[0]+1, self.ax[i].get_xlim()[1]-1], [cs[2], cs[2]], color=[0, 1, 0] ))
			
			# Flip the y axis so brains are the correct side up
			plt.gca().invert_yaxis()

			# Get rid of tick labels
			self.ax[i].set_xticks([])
			self.ax[i].set_yticks([])

			# Label the axes
			self.ax[i].set_xlabel(im_labels[i][0])
			self.ax[i].set_ylabel(im_labels[i][1])
			self.ax[i].axis(im_ranges[i])

		# Plot the maximum intensity projection
		self.ct_slice = 's' # Show sagittal MIP to start
		self.ax.append(self.fig.add_subplot(2,2,4))
		self.ax[3].set_axis_bgcolor('k')
		self.im.append(plt.imshow(np.nanmax(ct_data[cs[0]-15:cs[0]+15,:,:], axis=0).T, cmap=cm.gray, aspect='auto'))
		self.cursor.append(plt.plot([cs[1], cs[1]], [self.ax[3].get_ylim()[0]+1, self.ax[3].get_ylim()[1]-1], color=[0, 1, 0] ))
		self.cursor2.append(plt.plot([self.ax[3].get_xlim()[0]+1, self.ax[3].get_xlim()[1]-1], [cs[2], cs[2]], color=[0, 1, 0] ))
		self.ax[3].set_xticks([])
		self.ax[3].set_yticks([])
		plt.gca().invert_yaxis()
		self.ax[3].axis([0,self.imsz[1],0,self.imsz[2]])

		self.elec_im.append(plt.imshow(self.elec_data[cs[0],:,:].T, cmap=self.elec_colors, aspect='auto', alpha=1, vmin=0, vmax=17))
		plt.gcf().suptitle("Press 'n' to enter device name in console, press 'e' to add an electrode at crosshair, press 'h' for more options", fontsize=14)

		plt.tight_layout()
		plt.subplots_adjust(top=0.9)
		cid2 = self.fig.canvas.mpl_connect('scroll_event',self.on_scroll)
		cid3 = self.fig.canvas.mpl_connect('button_press_event',self.on_click)
		cid = self.fig.canvas.mpl_connect('key_press_event', self.on_key)
		#cid4 = self.fig.canvas.mpl_connect('key_release_event', self.on_key)

		plt.show()
		self.fig.canvas.draw()

	def on_key(self, event):
		''' 
		Executes when the user presses a key.  Potential key inputs are:

		Electrode adding:
		----
		n: enter the name of a new device (e.g. 'frontalgrid','hippocampaldepth')
		e: insert an electrode at the current green crosshair position
		u: remove electrode at the current crosshair position (can be thought of like "undo")

		Views:
		----
		s: sagittal view for maximum intensity projection at bottom right
		c: coronal view for maximum intensity projection at bottom right
		a: axial view for maximum intensity projection at bottom right
		pagedown/pageup: move by one slice in currently selected pane
		arrow up/arrow down: pan by one voxel in currently selected pane
		'''
		#print('You pressed', event.key)
		bb1=self.ax[0].get_position()
		bb2=self.ax[1].get_position()
		bb3=self.ax[2].get_position()
		bb4=self.ax[3].get_position()

		# Transform coordinates to figure coordinates
		fxy = self.fig.transFigure.inverted().transform((event.x, event.y))

		slice_num = []
		if bb1.contains(fxy[0],fxy[1]):
			slice_num = 0
		if bb2.contains(fxy[0],fxy[1]):
			slice_num = 1
		if bb3.contains(fxy[0],fxy[1]):
			slice_num = 2
		if bb4.contains(fxy[0],fxy[1]):
			slice_num = 3


		if event.key == 'escape':
			plt.close()

		if event.key == 't':
			# Toggle pial surface outline on and off
			self.pial_surf_on = not self.pial_surf_on

		if event.key == 'n':
			plt.gcf().suptitle("Enter electrode name in python console", fontsize=14)
			plt.gcf().canvas.draw()

			self.device_name = raw_input("Enter electrode name: ")
			plt.get_current_fig_manager().window.raise_()

			# If the device name is not in the list
			if self.device_name not in self.devices:
				self.devices.append(self.device_name)
				self.device_num = np.max(self.device_num)+1 # Find the next number 
			else:
				self.device_num = self.devices.index(self.device_name)
			
			plt.gcf().suptitle("Click on electrodes for device number %d, %s"%(self.device_num, self.device_name), fontsize=14)
			plt.gcf().canvas.draw()

			# If the device name is not in the list, start with electrode 0, or
			# load the electrode file if it exists and start with the next number
			# electrode
			if self.device_name not in self.elec_num:
				self.elec_num[self.device_name] = 0
				elecfile = os.path.join(self.subj_dir, 'elecs', 'individual_elecs', self.device_name+'.mat')
				if os.path.isfile(elecfile):
					emat = scipy.io.loadmat(elecfile)['elecmatrix']
					self.elecmatrix[self.device_name] = list(emat)
					print("Loading %s (if you wish to overwrite, remove this file before running)"%(elecfile))
					for elec in emat:
						self.current_slice = self.surfaceRAS_to_slice(elec[:3])
						self.add_electrode(add_to_file=False)
					print("Starting to mark electrode %d"%(self.elec_num[self.device_name]))
			self.update_legend()

		if event.key == 'h':
			# Show help 
			plt.gcf().suptitle("Help: 'n': name device, 'e': add electrode, 'u': remove electrode, 't': toggle pial surface, '3': show 3D view\nMaximum intensity projection views: 's': sagittal, 'c': coronal, 'a': axial\nScroll to zoom, arrows to pan, pgup/pgdown or click to go to slice", fontsize=12)

		if event.key == 'e':
			if self.device_name == '':
				plt.gcf().suptitle("Please name device with 'n' key before selecting electrode with 'e'", color='r', fontsize=14)
			else:
				self.add_electrode()

		if event.key == 'u':
			self.remove_electrode()
		
		if event.key == '3':
			self.launch_3D_viewer()

		# Maximum intensity projection in another dimension
		ct_slice = dict()
		ct_slice['s'] = 0
		ct_slice['c'] = 1
		ct_slice['a'] = 2
		if event.key == 's' or event.key == 'c' or event.key == 'a':
			self.ct_slice = event.key

		if slice_num != []:
			this_ax = self.ax[slice_num]

			# Scrolling through slices
			if event.key == 'pageup' or event.key == 'pagedown':
				if event.key == 'pagedown':
					sgn = -1
				else:
					sgn = 1
				if slice_num < 3:
					self.current_slice[slice_num] = self.current_slice[slice_num] + 1*sgn
				if slice_num == 3:
					self.current_slice[ct_slice[self.ct_slice]] = self.current_slice[ct_slice[self.ct_slice]] + 1*sgn

			# Panning left/right/up/down
			if event.key == 'up' or event.key == 'down' or event.key == 'left' or event.key == 'right':
				if event.key == 'up' or event.key == 'down':
					if event.key == 'up':
						sgn = -1
					else:
						sgn = 1
					ylims = this_ax.get_ylim()
					this_ax.set_ylim([ylims[0]+1*sgn, ylims[1]+1*sgn])
					
				elif event.key == 'left' or event.key == 'right':
					if event.key == 'right':
						sgn = -1
					else:
						sgn = 1
					xlims = this_ax.get_xlim()
					this_ax.set_xlim(xlims[0]+1*sgn, xlims[1]+1*sgn)

		# Draw the figure
		self.update_figure_data(ax_clicked=slice_num)
		plt.gcf().canvas.draw()

	def on_scroll(self, event):
		''' Use mouse scroll wheel to zoom.  Scroll down zooms in, scroll up zooms out.
		'''
		stepsz = 5.

		xstep = event.step*stepsz
		ystep = event.step*stepsz

		for a in np.arange(4):
			this_ax = self.ax[a]
			xlims = this_ax.get_xlim()
			ylims = this_ax.get_ylim()
			if xlims[0] + xstep > xlims[1] - xstep:
				this_ax.set_xlim(xlims[0], xlims[1])
			else:
				this_ax.set_xlim(xlims[0]+xstep, xlims[1] - xstep)
			if ylims[0] + ystep > ylims[1] - ystep:
				this_ax.set_ylim(ylims[0], ylims[1])
			else:
				this_ax.set_ylim(ylims[0]+ystep, ylims[1] - ystep)
			self.cursor[a][0].set_ydata ([self.ax[a].get_ylim()]) 
			self.cursor2[a][0].set_xdata([self.ax[a].get_xlim()])

		plt.gcf().canvas.draw()

	def on_click(self, event):
		'''
		Executes on mouse click events -- moves appropriate subplot axes to (x,y,z)
		view on MRI and CT views.
		'''

		#print('You scrolled %d steps at x: %d, y: %d', event.step, event.x, event.y)
		# Get the bounding box for each of the subplots
		bb1=self.ax[0].get_position()
		bb2=self.ax[1].get_position()
		bb3=self.ax[2].get_position()
		bb4=self.ax[3].get_position()

		#print event.xdata, event.ydata
		# Transform coordinates to figure coordinates
		fxy = self.fig.transFigure.inverted().transform((event.x, event.y))
		
		x = np.int(np.round(event.xdata))
		y = np.int(np.round(event.ydata))

		# If you clicked the first subplot
		if bb1.contains(fxy[0],fxy[1]):
			self.current_slice[1] = event.xdata
			self.current_slice[2] = event.ydata
			ax_num = 0
			
		# If you clicked the second subplot
		elif bb2.contains(fxy[0],fxy[1]):
			self.current_slice[0] = event.xdata
			self.current_slice[2] = event.ydata
			ax_num = 1

		# If you clicked the third subplot
		elif bb3.contains(fxy[0],fxy[1]):
			self.current_slice[0] = event.xdata
			self.current_slice[1] = event.ydata
			ax_num = 2

		# If you clicked the third subplot
		elif bb4.contains(fxy[0],fxy[1]):
			if self.ct_slice == 's':
				self.current_slice[1] = event.xdata
				self.current_slice[2] = event.ydata
			elif self.ct_slice == 'c':
				self.current_slice[0] = event.xdata
				self.current_slice[2] = event.ydata
			elif self.ct_slice == 'a':
				self.current_slice[0] = event.xdata
				self.current_slice[1] = event.ydata
			ax_num = 3

		self.update_figure_data(ax_clicked=ax_num)

		#print("Current slice: %3.2f %3.2f %3.2f"%(self.current_slice[0], self.current_slice[1], self.current_slice[2]))
		plt.gcf().canvas.draw()

	def update_figure_data(self, ax_clicked=None):
		'''
		Updates all four subplots based on the crosshair position (self.current_slice)
		The subplots (in order) are the sagittal view, coronal view, and axial view,
		followed by the maximum intensity projection of the CT scan (in the user
		specified view, which is sagittal by default)
		'''
		cs = np.round(self.current_slice).astype(np.int) # Make integer for indexing the volume
		self.im[0].set_data(self.img_data[cs[0],:,:].T)
		self.im[1].set_data(self.img_data[:,cs[1],:].T)
		self.im[2].set_data(self.img_data[:,:,cs[2]].T)

		# Show the maximum intensity projection for +/- 15 slices
		# Sagittal view
		if self.ct_slice == 's':
			self.im[3].set_data(np.nanmax(self.ct_data[cs[0]-15:cs[0]+15,:,:], axis=0).T)
		# Coronal view
		elif self.ct_slice == 'c':
			self.im[3].set_data(np.nanmax(self.ct_data[:,cs[1]-15:cs[1]+15,:], axis=1).T)
		# Axial view
		elif self.ct_slice == 'a':
			self.im[3].set_data(np.nanmax(self.ct_data[:,:,cs[2]-15:cs[2]+15], axis=2).T)

		# Show the CT data in the sagittal, coronal, and axial views
		self.ct_im[0].set_data(self.ct_data[cs[0],:,:].T)
		self.ct_im[1].set_data(self.ct_data[:,cs[1],:].T)
		self.ct_im[2].set_data(self.ct_data[:,:,cs[2]].T)

		# Show the pial surface data in the sagittal, coronal, and axial views
		for a in np.arange(3):
			if self.contour[a]:
				try:
					for coll in self.pial_im[a].collections:
						coll.remove()
				except:
					pass
		
		if self.pial_surf_on:
			if np.any(self.pial_data[cs[0],:,:]):
				self.pial_im[0] = self.ax[0].contour(self.pial_data[cs[0],:,:].T, linewidths=0.5, colors = 'y')
				self.contour[0] = True
			else:
				self.contour[0] = False
			if np.any(self.pial_data[:,cs[1],:]):
				self.pial_im[1] = self.ax[1].contour(self.pial_data[:,cs[1],:].T, linewidths=0.5, colors = 'y')
				self.contour[1] = True
			else:
				self.contour[1] = False
			if np.any(self.pial_data[:,:,cs[2]]):
				self.pial_im[2] = self.ax[2].contour(self.pial_data[:,:,cs[2]].T, linewidths=0.5, colors = 'y')
				self.contour[2] = True
			else:
				self.contour[2] = False
		
		# Show the electrode volume data in the sagittal, coronal, and axial views
		self.elec_im[0].set_data(self.elec_data[cs[0],:,:].T)
		self.elec_im[1].set_data(self.elec_data[:,cs[1],:].T)
		self.elec_im[2].set_data(self.elec_data[:,:,cs[2]].T)

		# Slow the electrodes on the maximum intensity projection, 
		# make sure the correct slices are shown based on which orientation
		# we're using
		if self.ct_slice == 's':
			self.elec_im[3].set_data(self.elec_data[cs[0],:,:].T)
			self.cursor[3][0].set_xdata ([self.current_slice[1], self.current_slice[1]]) 
			self.cursor2[3][0].set_ydata([self.current_slice[2], self.current_slice[2]])
		elif self.ct_slice == 'c':
			self.elec_im[3].set_data(self.elec_data[:,cs[1],:].T)
			self.cursor[3][0].set_xdata ([self.current_slice[0], self.current_slice[0]]) 
			self.cursor2[3][0].set_ydata([self.current_slice[2], self.current_slice[2]])
		elif self.ct_slice == 'a':
			self.elec_im[3].set_data(self.elec_data[:,:,cs[2]].T)
			self.cursor[3][0].set_xdata ([self.current_slice[0], self.current_slice[0]]) 
			self.cursor2[3][0].set_ydata([self.current_slice[1], self.current_slice[1]])

		# Set the crosshairs for the sagittal (0), coronal (1), and axial (2) views
		self.cursor[0][0].set_xdata ([self.current_slice[1], self.current_slice[1]])
		self.cursor[0][0].set_ydata ([self.ax[0].get_ylim()])
		self.cursor2[0][0].set_ydata([self.current_slice[2], self.current_slice[2]])
		self.cursor2[0][0].set_xdata ([self.ax[0].get_xlim()])
		self.cursor[1][0].set_xdata ([self.current_slice[0], self.current_slice[0]])
		self.cursor[1][0].set_ydata ([self.ax[1].get_ylim()])
		self.cursor2[1][0].set_ydata([self.current_slice[2], self.current_slice[2]])
		self.cursor2[1][0].set_xdata ([self.ax[1].get_xlim()])
		self.cursor[2][0].set_xdata ([self.current_slice[0], self.current_slice[0]])
		self.cursor[2][0].set_ydata ([self.ax[2].get_ylim()])
		self.cursor2[2][0].set_ydata([self.current_slice[1], self.current_slice[1]])
		self.cursor2[2][0].set_xdata ([self.ax[2].get_xlim()])

		# Re-center the plots at the crosshair location
		for a in np.arange(4):
			# Only re-center plots that you didn't click on (since it's annoying
			# when the plot that you just clicked moves...)
			if a!=ax_clicked:
				xlims = self.ax[a].get_xlim()
				xax_range = xlims[1]-xlims[0]
				ylims = self.ax[a].get_ylim()
				yax_range = ylims[1]-ylims[0]
				center_pt_x = self.cursor[a][0].get_xdata()[0]
				center_pt_y = self.cursor2[a][0].get_ydata()[0]
				self.ax[a].set_xlim(center_pt_x - xax_range/2., center_pt_x + xax_range/2.)
				self.ax[a].set_ylim(center_pt_y - yax_range/2., center_pt_y + yax_range/2.)
				self.cursor[a][0].set_ydata ([self.ax[a].get_ylim()]) 
				self.cursor2[a][0].set_xdata([self.ax[a].get_xlim()])

	def add_electrode(self, add_to_file = True):
		'''
		Add an electrode at the current crosshair point. 
		'''

		# Make the current slice into an integer for indexing the volume
		cs = np.round(self.current_slice).astype(np.int) 

		# Create a sphere centered around the current point as a binary matrix
		radius = 2
		r2 = np.arange(-radius, radius+1)**2
		dist2 = r2[:,None,None]+r2[:,None]+r2
		bin_mat = np.array(dist2<=radius**2, dtype=np.float)
		bin_mat[bin_mat==0] = np.nan
		
		# The sphere part of the binary matrix will have a value that
		# increments with device number so that different devices
		# will show up in different colors
		bin_mat = bin_mat+self.device_num-2
		self.bin_mat = bin_mat
		
		# Set the electrode data volume for this bounding box (add the
		# sphere to the electrode "volume" so it shows up in the brain
		# plots)
		self.elec_data[cs[0]-radius:cs[0]+radius+1, cs[1]-radius:cs[1]+radius+1, cs[2]-radius:cs[2]+radius+1] = bin_mat

		self.elec_im[0].set_data(self.elec_data[cs[0],:,:].T)
		self.elec_im[0].set_data(self.elec_data[:,cs[1],:].T)
		self.elec_im[0].set_data(self.elec_data[:,:,cs[2]].T)

		# As displayed, these coordinates are LSP, and we want RAS,
		# so we do that here
		elec = self.slice_to_surfaceRAS()

		# Find whether this device already exists and append to it if it does, or
		# initialize a new one
		if self.device_name not in self.elecmatrix:
			# Initialize the electrode matrix
			self.elecmatrix[self.device_name] = []
		self.elecmatrix[self.device_name].append(elec)

		# Add the electrode to the file (we wouldn't want to 
		# do this if we are displaying previously clicked electrodes)
		if add_to_file:
			elecfile = os.path.join(self.subj_dir, 'elecs', 'individual_elecs', self.device_name+'.mat')
			scipy.io.savemat(elecfile, {'elecmatrix': np.array(self.elecmatrix[self.device_name])})

		plt.gcf().suptitle('%s e%d surface RAS = [%3.3f, %3.3f, %3.3f]'%(self.device_name, self.elec_num[self.device_name], elec[0], elec[1], elec[2]), fontsize=14)
		
		self.elec_num[self.device_name] += 1
		#print("Voxel CRS: %3.3f, %3.3f, %3.3f"%(self.current_slice[0], self.current_slice[1], self.current_slice[2]))
		#print("RAS coordinate: %3.3f, %3.3f, %3.3f"%(elec[0], elec[1], elec[2]))

	def remove_electrode(self):
		'''
		Remove the electrode at the current crosshair point. 
		'''
		cs = self.current_slice
		if self.bin_mat != '':
			self.bin_mat = ''
			
			# Remove the electrode from elecmatrix
			self.elecmatrix[self.device_name].pop()

			# Save the electrode matrix
			elecfile = os.path.join(self.subj_dir, 'elecs', 'individual_elecs', self.device_name+'.mat')
			scipy.io.savemat(elecfile, {'elecmatrix': np.array(self.elecmatrix[self.device_name])})

			# Remove the electrode from the volume display
			self.elec_data[cs[0]-radius:cs[0]+radius+1, cs[1]-radius:cs[1]+radius+1, cs[2]-radius:cs[2]+radius+1] = np.nan

			self.elec_im[0].set_data(self.elec_data[cs[0],:,:].T)
			self.elec_im[0].set_data(self.elec_data[:,cs[1],:].T)
			self.elec_im[0].set_data(self.elec_data[:,:,cs[2]].T)

	def slice_to_surfaceRAS(self, coord = None):
		'''
		Convert slice coordinate from the viewer to surface RAS
		'''
		if coord is None:
			coord = self.current_slice

		elec_CRS = np.hstack((self.imsz[0] - coord[0],
							  self.imsz[2] - coord[2],
							  coord[1], 1))
		
		# Convert CRS to RAS
		elec = np.dot(self.fsVox2RAS, elec_CRS.transpose()).transpose()
		elec = elec[:3]

		return elec

	def surfaceRAS_to_slice(self, elec):
		'''
		Convert surface RAS to coordinate to be used in the viewer
		'''
		elec = np.hstack((elec, 1))

		# Convert CRS to RAS
		elec_CRS = np.dot(np.linalg.inv(self.fsVox2RAS), elec.transpose()).transpose()

		print(elec_CRS)
		coord = np.hstack((self.imsz[0] - elec_CRS[0],
						   elec_CRS[2],
						   self.imsz[1] - elec_CRS[1]))
		
		return coord

	def update_legend(self, vmax=17.):
		''' 
		Update the legend with the electrode devices.
		'''
		self.legend_handles = []
		for i in self.devices:
			QtCore.pyqtRemoveInputHook()
			cmap = self.elec_colors
			num = self.devices.index(i)
			c = cmap(num/vmax)
			color_patch = mpatches.Patch(color=c, label=i)
			self.legend_handles.append(color_patch)
			plt.legend(handles=self.legend_handles, loc='upper right', fontsize='x-small')

	def launch_3D_viewer(self):
		'''
		Launch 3D viewer showing position of identified electrodes in 3D space
		'''
		from plotting.ctmr_brain_plot import ctmr_gauss_plot as ctmr_gauss_plot
		from plotting.ctmr_brain_plot import el_add as el_add
		
		# Get appropriate hemisphere
		pial = scipy.io.loadmat(os.path.join(self.subj_dir, 'Meshes', self.hem+'_pial_trivert.mat'))
		ctmr_gauss_plot(pial['tri'], pial['vert'], opacity=0.8)
		#rh = scipy.io.loadmat(os.path.join(self.subj_dir, 'Meshes', 'rh_pial_trivert.mat'))
		#ctmr_gauss_plot(rh['tri'], rh['vert'], opacity=0.8, new_fig=False)

		# Plot the electrodes we have so far
		vmax = 17.
		for i, dev in enumerate(self.devices):
			elecfile = os.path.join(self.subj_dir, 'elecs', 'individual_elecs', dev+'.mat')
			e = scipy.io.loadmat(elecfile)['elecmatrix']
			num = self.devices.index(dev)
			c = self.elec_colors(num/vmax)
			el_add(e, color=tuple(c[:3]), msize=4, numbers=1+np.arange(e.shape[0]))


if __name__ == '__main__':
	app = QtGui.QApplication([])
	path_to_this_func = os.path.dirname(os.path.realpath(__file__))
	app.setWindowIcon(QtGui.QIcon(os.path.join(path_to_this_func, 'icons','leftbrain.png')))
	subj_dir = sys.argv[1]
	hem = sys.argv[2]
	e = electrode_picker(subj_dir = subj_dir, hem = hem)



import sys
import numpy as np
import nibabel as nib
from scipy.signal import convolve
from scipy.ndimage.morphology import grey_closing, generate_binary_structure
from mne import write_surface
from mcubes import marching_cubes

def make_outer_surf(orig_pial, image, radius, outfile):
    '''
    Make outer surface based on a pial volume and radius,
    write to surface in outfile.

    Args:
        orig_pial: pial surface (e.g. lh.pial)
        image: filled lh or rh pial image (e.g. lh.pial.filled.mgz)
        radius: radius for smoothing (currently ignored)
        outfile: surface file to write data to

    Original code from ielu (https://github.com/aestrivex/ielu)
    '''
    
    #radius information is currently ignored
    #it is a little tougher to deal with the morphology in python

    pial_surf = nib.freesurfer.read_geometry(orig_pial, read_metadata=True)
    volume_info = pial_surf[2]

    fill = nib.load( image )
    filld = fill.get_data()
    filld[filld==1] = 255

    gaussian = np.ones((2,2))*.25

    image_f = np.zeros((256,256,256))

    for slice in xrange(256):
        temp = filld[:,:,slice]
        image_f[:,:,slice] = convolve(temp, gaussian, 'same')

    image2 = np.zeros((256,256,256))
    image2[np.where(image_f <= 25)] = 0
    image2[np.where(image_f > 25)] = 255

    strel15 = generate_binary_structure(3, 1)

    BW2 = grey_closing(image2, structure=strel15)
    thresh = np.max(BW2)/2
    BW2[np.where(BW2 <= thresh)] = 0
    BW2[np.where(BW2 > thresh)] = 255

    v, f = marching_cubes(BW2, 100)

    v2 = np.transpose(
             np.vstack( ( 128 - v[:,0],
                          v[:,2] - 128,
                          128 - v[:,1], )))
    
    write_surface(outfile, v2, f, volume_info=volume_info)

if __name__=='__main__':

    if not len(sys.argv) == 4:
        raise ValueError("Usage error: please provide the following arguments:\n(1) original pial surface (used for metadata), (2) filled volume,\n(3) diameter integral, (4) output file ")

    make_outer_surf( sys.argv[1], None, sys.argv[3] )

    




#plot_recon_anatomy.py
#
# Liberty Hamilton 2016
#

import mayavi
import scipy.io
import ctmr_brain_plot
from ../SupplementalFiles import FS_colorLUT
import numpy as np
import os

fs_dir = os.environ['SUBJECTS_DIR']

def plot_recon_anatomy(patient):
	subj = patient.subj
	hem = patient.hem
	a = scipy.io.loadmat('%s/%s/Meshes/%s_pial_trivert.mat'%(fs_dir, subj, hem))
	e = scipy.io.loadmat('%s/%s/elecs/TDT_elecs_all.mat'%(fs_dir, subj))

	# Plot the pial surface
	mesh, mlab = ctmr_brain_plot.ctmr_gauss_plot(a['tri'], a['vert'], color=(0.8, 0.8, 0.8))
	
	# Add the electrodes, colored by anatomical region
	elec_colors = np.zeros((e['elecmatrix'].shape[0], e['elecmatrix'].shape[1]))

	# Import freesurfer color lookup table as a dictionary
	cmap = FS_colorLUT.get_lut()

	# Make a list of electrode numbers
	elec_numbers = np.arange(e['elecmatrix'].shape[0])+1

	# Find all the unique brain areas in this subject
	brain_areas = np.unique(e['anatomy'][:,3])

	# Loop through unique brain areas and plot the electrodes in each brain area
	for b in brain_areas:
	    # Add relevant extra information to the label if needed for the color LUT
	    this_label = b[0]
	    if b[0][0:3]!='ctx' and b[0][0:4] != 'Left' and b[0][0:5] != 'Right' and b[0][0:5] != 'Brain' and b[0] != 'Unknown':
	        this_label = 'ctx-%s-%s'%(hem, b[0])
	        print(this_label)
	   	
	    if this_label != '':
	        el_color = np.array(cmap[this_label])/255.
	        ctmr_brain_plot.el_add(np.atleast_2d(e['elecmatrix'][e['anatomy'][:,3]==b,:]), 
	        					   color=tuple(el_color), numbers=elec_numbers[e['anatomy'][:,3]==b])
	mlab.show()        
	return mesh, mlab



import numpy as np

def project_electrodes_anydirection(tri, vert, elecmatrix, proj_direction):
    ''' 
    Projects electrode locations onto the convex hull of a cortical surface.
    This allows for the electrode locations to wrap smoothly around the
    cortical surface instead of being outside the brain or going into the
    sulci.

     Inputs: cortex:       cortical surface struct (with fields cortex.vert
                           and cortex.tri)
             elecmatrix:   a [nchans] x 3 electrode position array of locations
                           of electrodes on the cortical surface (don't use
                           this script for depth electrodes)
             proj_direction:    a string with value
                           'lh','rh','top','bottom','front','back' depending
                           on where you want electrodes to project to

     Output: elecs_proj:   A [nchans] x 3 position array of locations of
                           electrodes that have been projected to the convex
                           hull of the cortical surface.

     Written Dec. 2014 by Liberty Hamilton

     See also: TRIANGLERAYINTERSECTION, written by Jarek Tuszynski, available
     at http://www.mathworks.com/matlabcentral/fileexchange/33073-triangle-ray-intersection
    '''

    # This flag is for calculating light direction as well as direction of line
    # of intersection
    if type(proj_direction) is list:
        direction = proj_direction
    elif proj_direction=='lh':
    	direction = [1000, 0, 0] # IMPORTANT! This is not a point.
    	# This is the direction to add to the
    	# "orig" coordinate to get the line you
    	# wish to intersect
    elif proj_direction=='rh':
    	direction = [-1000, 0, 0]

    vert1 = vert[tri[:,0],:]
    vert2 = vert[tri[:,1],:]
    vert3 = vert[tri[:,2],:]

    elecs_proj = np.zeros(elecmatrix.shape)
    elec_intersect = np.zeros((elecmatrix.shape[0],1))
    for i in range(elecmatrix.shape[0]): # Loop through all electrodes  
        # Define a line with original (off-brain) electrode position as the
        # origin and pointing in a direction parallel to the x axis (so the
        # grid will only move left-right)
        orig = elecmatrix[i,:]
        # Calculate the intersection of the electrode with the convex hull mesh
        elec_intersect,_,_,_,xcoor = TriangleRayIntersection(orig, direction, vert1, vert2, vert3)
        #print xcoor[elec_intersect,:] , elec_intersect.sum(), np.where(elec_intersect>0), xcoor.shape
        # If there is more than one intersection of this ray with the cortical
        # mesh, only get the most negative coordinate for the left hemisphere,
        # or only the most positive coordinate for the right hemisphere (this
        # is for surface electrodes only!)

        if np.sum(elec_intersect)>1:
            xctmp = xcoor[elec_intersect,:] # all intersecting coordinates
            # #code block that selects based on lh or rh
            # sorted_xctmp = np.sort(xctmp,axis=0)
            # if proj_direction=='lh':
            #     x = sorted_xctmp[0,:]
            # else: # right hemisphere
            #     x = sorted_xctmp[-1,:]

            #instead, select from competing intersection coordinates using least euclidean distance metric
            x = xctmp[np.argmin(np.sum(np.abs(orig-xctmp)**2,axis=1)**(1./2)),:]
            elecs_proj[i,:] = x
        elif np.sum(elec_intersect)==1:
            x = xcoor[elec_intersect,:]
        else:
            x = np.nan
        elecs_proj[i,:] = x

    return elecs_proj

    # Not currently implemented: (this could replace the convex hull step)
    #[V,S]=alphavol(cortex.vert,20,1) # May consider using this for electrodes
    # on bottom surface or concave surfaces... need to develop!!
    #figure trisurf(S.bnd, cortex.vert(:,1), cortex.vert(:,2), cortex.vert(:,3))

def TriangleRayIntersection (orig, direction, vert0, vert1, vert2, planeType='two sided', border='normal', eps=1e-5, fullReturn=False):
    # TRIANGLERAYINTERSECTION Ray/triangle intersection.
    #    INTERSECT = TriangleRayIntersection(ORIG, DIRECTION, VERT1, VERT2, VERT3) 
    #      calculates ray/triangle intersections using the algorithm proposed
    #      BY Moller and Trumbore (1997), implemented as highly vectorized
    #      MATLAB code. The ray starts at ORIG and points toward DIRECTION. The 
    #      triangle is defined by vertix points: VERT1, VERT2, VERT3. All input  
    #      arrays are in Nx3 or 1x3 format, where N is number of triangles or 
    #      rays.
    # 
    #   [INTERSECT, T, U, V, XCOOR] = TriangleRayIntersection(...) 
    #     Returns:
    #     * Intersect - boolean array of length N informing which line and
    #                 triangle pair intersect
    #     * t   - distance from the ray origin to the intersection point in 
    #             units of |dir|. Provided only for line/triangle pair that 
    #             intersect unless 'fullReturn' parameter is true.
    #     * u,v - barycentric coordinates of the intersection point 
    #     * xcoor - carthesian coordinates of the intersection point
    #
    #   TriangleRayIntersection(...,'param','value','param','value'...) allows
    #    additional param/value pairs to be used. Allowed parameters:
    #    * planeType - 'one sided' or 'two sided' (default) - how to treat
    #        triangles. In 'one sided' version only intersections in single
    #        direction are counted and intersections with back facing
    #           tringles are ignored
    #    * border - controls border handling:
    #        - 'normal'(default) border - triangle is exactly as defined. 
    #           Intersections with border points can be easily lost due to
    #           rounding errors. 
    #        - 'inclusive' border - triangle is marginally larger.
    #           Intersections with border points are always captured but can
    #           lead to double counting when working with surfaces.
    #        - 'exclusive' border - triangle is marginally smaller. 
    #           Intersections with border points are not captured and can
    #           lead to under-counting when working with surfaces.
    #    * epsilon - (default = 1e-5) controls border size
    #    * fullReturn - (default = false) controls returned variables t, u, v, 
    #        and xcoor
    #
    # ALGORITHM:
    #  Function solves
    #        |t|
    #    M * |u| = (o-v0)
    #        |v|
    #  for [t; u; v] where M = [-d, v1-v0, v2-v0]. u,v are barycentric coordinates
    #  and t - the distance from the ray origin in |d| units
    #  ray/triangle intersect if u>=0, v>=0 and u+v<=1
    #
    # NOTE:
    #  The algorithm is able to solve several types of problems:
    #  * many faces / single ray  intersection
    #  * one  face  / many   rays intersection
    #  * one  face  / one    ray  intersection
    #  * many faces / many   rays intersection
    #  In order to allow that to happen all imput arrays are expected in Nx3
    #  format, where N is number of vertices or rays. In most cases number of
    #  vertices is different than number of rays, so one of the imputs will
    #  have to be cloned to have the right size. Use "repmat(A,size(B,1),1)".
    #
    # Based on:
    #  *"Fast, minimum storage ray-triangle intersection". Tomas Moller and
    #    Ben Trumbore. Journal of Graphics Tools, 2(1):21--28, 1997.
    #    http://www.graphics.cornell.edu/pubs/1997/MT97.pdf
    #  * http://fileadmin.cs.lth.se/cs/Personal/Tomas_Akenine-Moller/raytri/
    #  * http://fileadmin.cs.lth.se/cs/Personal/Tomas_Akenine-Moller/raytri/raytri.c
    #
    # Author:
    #    Jarek Tuszynski (jaroslaw.w.tuszynski@leidos.com)
    #
    # License: BSD license (http://en.wikipedia.org/wiki/BSD_licenses)
    # 
    # Converted to python by Liberty Hamilton 2017

    # In case of single points clone them to the same size as the rest
    orig, direction, vert0, vert1, vert2 = np.array(orig),np.array(direction),np.array(vert0),np.array(vert1),np.array(vert2)

    N = max(orig.shape[0], direction.shape[0], vert0.shape[0], vert1.shape[0],vert2.shape[0])
    if orig.shape[0]==3:
        orig = np.tile(orig,(N,1))
    if direction.shape[0]==3:
        direction = np.tile(direction,(N,1))
    #if (size(orig ,1)==1 && N>1 && size(orig ,2)==3), orig  = repmat(orig , N, 1); end
    #if (size(direction  ,1)==1 && N>1 && size(direction  ,2)==3), direction   = repmat(direction  , N, 1); end
    #if (size(vert0,1)==1 && N>1 && size(vert0,2)==3), vert0 = repmat(vert0, N, 1); end
    #if (size(vert1,1)==1 && N>1 && size(vert1,2)==3), vert1 = repmat(vert1, N, 1); end
    #if (size(vert2,1)==1 && N>1 && size(vert2,2)==3), vert2 = repmat(vert2, N, 1); end

    # Check if all the sizes match
    # SameSize = (any(size(orig)==size(vert0)) && ...
    #   any(size(orig)==size(vert1)) && ...
    #   any(size(orig)==size(vert2)) && ...
    #   any(size(orig)==size(dir  )) );
    # assert(SameSize && size(orig,2)==3, ...
    #   'All input vectors have to be in Nx3 format.');

    # Set up border parameter
    if border=='normal':
        zero=0.0
    elif border=='inclusive':
        zero=eps
    elif border=='exclusive':
        zero=-eps
    else:
        print("Using 'normal' border parameter")
        zero=0.0

    # initialize default output
    intersect = np.zeros((orig.shape[0],), dtype=bool) # by default there are no intersections
    t = np.inf + np.zeros((orig.shape[0],))
    u = t
    v = t 

    # Find faces parallel to the ray
    edge1 = vert1 - vert0          # find vectors for two edges sharing vert0
    edge2 = vert2 - vert0
    tvec  = orig - vert0          # vector from vert0 to ray origin
    pvec  = np.cross(direction, edge2) # begin calculating determinant - also used to calculate U parameter
    det   = np.sum(edge1*pvec, axis=1)   # determinant of the matrix M = dot(edge1,pvec)


    if planeType=='two sided':    # treats triangles as two sided
        angleOK = np.abs(det)>eps  # if determinant is near zero then ray lies in the plane of the triangle
    elif planeType=='one sided':  # treats triangles as one sided
        angleOK = det>eps
    else:
        print("Using two sided plane")
        angleOK = np.abs(det)>eps

    # if all parallel than no intersections
    if not np.any(angleOK):
        intersect = False

    # Different behavior depending on one or two sided triangles
    det[np.invert(angleOK)] = np.nan              # change to avoid division by zero
    u    = np.sum(tvec*pvec, axis=1)/det    # 1st barycentric coordinate

    if fullReturn:
        # calculate all variables for all line/triangle pairs
        qvec = np.cross(tvec, edge1)    # prepare to test V parameter
        v    = np.sum(direction  *qvec,axis=1)/det   # 2nd barycentric coordinate
        t    = np.sum(edge2*qvec,axis=1)/det   # 'position on the line' coordinate
        # test if line/plane intersection is within the triangle
        ok   = (angleOK) & (u>=-zero) & (v>=-zero) & ((u+v)<=(1.0+zero))
    else:
        # limit some calculations only to line/triangle pairs where it makes
        # a difference. It is tempting to try to push this concept of
        # limiting the number of calculations to only the necessary to "u"
        # and "t" but that produces slower code
        v = np.nan+np.zeros((u.shape[0], ))
        t = v
        ok = (angleOK) & (u>=-zero) & (u<=1.0+zero) # mask
        
        # if all line/plane intersections are outside the triangle than no intersections
        if not any(ok):
            intersect = ok

        qvec = np.cross(tvec[ok,:], edge1[ok,:]) # prepare to test V parameter
        v[ok] = np.sum(direction[ok,:]*qvec,axis=1) / det[ok] # 2nd barycentric coordinate

        # test if line/plane intersection is within the triangle
        ok = (ok) & (v>=-zero) & ((u+v)<=(1.0+zero))


    # Test where along the line the line/plane intersection occurs
    intersect = ok

    # calculate intersection coordinates if requested
    xcoor = np.nan+np.zeros((orig.shape))
    ok = intersect
    #xcoor = vert0[ok,:] + edge1[ok,:]*u[ok] + edge2[ok,:]*v[ok]
    for o in np.where(ok)[0]:
        xcoor[o,:] = (vert0[o,:] + edge1[o,:]*u[o] + edge2[o,:]*v[o])

    return np.array(intersect), np.array(t), np.array(u), np.array(v), xcoor
